from save_to_db.core.item import Item
from save_to_db.utils.test_base import TestBase



class TestAdapter(TestBase):
    
    ModelGeneralOne = None
    ModelGeneralTwo = None
    
    
    def test_multiple_same_models_returned(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = ['f_integer']
            getters = ['two_x_x', 'f_string']
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            creators = ['f_integer']
            getters = ['f_string']
        
        persister = self.persister
        
        item_one = ItemGeneralOne(f_integer='1', f_string='str-1')
        item_two_1 = ItemGeneralTwo(f_integer='2', f_string='str-2')
        item_two_2 = ItemGeneralTwo(f_integer='3', f_string='str-3')
        # through x-to-many relation same model can be loaded twice
        item_one['two_x_x'].add(item_two_1, item_two_2)
        
        def ensure_saved_models():
            models_one = self.get_all_models(self.ModelGeneralOne)
            models_two = self.get_all_models(self.ModelGeneralTwo,
                                             sort_key=lambda x: x.f_integer)
            self.assertEqual(len(models_one), 1)
            self.assertEqual(len(models_two), 2)
            model_one = self.get_all_models(self.ModelGeneralOne)[0]
            model_two_1 = self.get_all_models(self.ModelGeneralTwo)[0]
            model_two_2 = self.get_all_models(self.ModelGeneralTwo)[1]
            
            self.assertEqual(model_one.f_integer, 1)
            self.assertEqual(model_one.f_string, 'str-1')
            related_two = self.get_related_x_to_many(model_one, 'two_x_x')
            related_two.sort(key=lambda x: x.f_integer)
            self.assertEqual(len(related_two), 2)
            self.assertEqual(model_two_1, related_two[0])
            self.assertEqual(model_two_2, related_two[1])
            
            self.assertEqual(model_two_1.f_integer, 2)
            self.assertEqual(model_two_1.f_string, 'str-2')
            related_one = self.get_related_x_to_many(model_two_1,
                                                            'one_x_x')
            self.assertEqual(len(related_one), 1)
            self.assertEqual(model_one, related_one[0])
            
            self.assertEqual(model_two_2.f_integer, 3)
            self.assertEqual(model_two_2.f_string, 'str-3')
            related_one = self.get_related_x_to_many(model_two_2, 'one_x_x')
            self.assertEqual(len(related_one), 1)
            self.assertEqual(model_one, related_one[0])
        
        persister.persist(item_one)
        ensure_saved_models()
        
        # saving models second time might cause same model to be returned
        # for ['two_x_x', 'f_string'] getters
        persister.persist(item_one)
        ensure_saved_models()
    
    
    def test_no_item_returned_with_fkey(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = [['f_integer', 'two_1_1']]
            getters = [['f_integer', 'two_1_1']]
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            creators = ['f_integer']
            getters = ['f_integer']
            
        persister = self.persister
        db_adapter = persister.db_adapter
        
        item_one = ItemGeneralOne(f_integer=1)
        item_one['two_1_1'](f_integer=2)
        
        persister.persist(item_one)
        models_one = self.get_all_models(self.ModelGeneralOne)
        self.assertEqual(len(models_one), 1)
        model_one = models_one[0]
        self.assertEqual(model_one.f_integer, 1)
        self.assertIsNotNone(model_one.two_1_1)
        self.assertEqual(model_one.two_1_1.f_integer, 2)
        
        item_one = ItemGeneralOne(f_integer=1)
        item_one['two_1_1'](f_integer=2)
        items_and_fkeys = [[item_one, {}],]
        got_models = db_adapter.get(items_and_fkeys)
        self.assertFalse(bool(got_models),
                         'Adapter returned models that have relations in '
                         'getter when empty `fkey`. Probably adapter '
                         'just returned all models from DB.')
    
    
    def test_related_x_to_many_contains(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = ['f_integer']
            getters = ['f_integer']
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            creators = ['f_integer']
            getters = ['f_integer']
        
        persister = self.persister
        db_adapter = persister.db_adapter
        
        for f_key in ['two_1_x', 'two_x_x']:
            item_one_1 = ItemGeneralOne(f_integer='1000')
            item_one_1[f_key].gen(f_integer='1001')
            item_one_1[f_key].gen(f_integer='1002')
            
            item_one_2 = ItemGeneralOne(f_integer='2000')
            item_one_2[f_key].gen(f_integer='2001')
            item_one_2[f_key].gen(f_integer='2002')
            
            persister.persist(item_one_1)
            persister.persist(item_one_2)
            
            sort_func = lambda model: model.f_integer
            
            model_ones = self.get_all_models(self.ModelGeneralOne,
                                             sort_key=sort_func)
            self.assertEqual(len(model_ones), 2)
            model_one_1000, model_one_2000 = model_ones
            
            models_two_1 = self.get_related_x_to_many(model_one_1000, f_key)
            self.assertEqual(len(models_two_1), 2)
            model_two_1001, model_two_1002 = models_two_1
            
            models_two_2 = self.get_related_x_to_many(model_one_2000, f_key)
            self.assertEqual(len(models_two_2), 2)
            model_two_2001, model_two_2002 = models_two_2
            
            self.assertEqual(model_one_1000.f_integer, 1000)
            self.assertEqual(model_one_2000.f_integer, 2000)
            self.assertEqual(model_two_1001.f_integer, 1001)
            self.assertEqual(model_two_1002.f_integer, 1002)
            self.assertEqual(model_two_2001.f_integer, 2001)
            self.assertEqual(model_two_2002.f_integer, 2002)
            
            # the testing
            models_two = db_adapter.related_x_to_many_contains(
                model_one_1000, f_key, [model_two_1001, model_two_1002,
                                        model_two_2001, model_two_2002])
            self.assertEqual(len(models_two), 2)
            models_two.sort(key=sort_func)
            self.assertEqual(models_two[0].f_integer, 1001)
            self.assertEqual(models_two[1].f_integer, 1002)
            # exectly the same models returned
            self.assertIs(model_two_1001, models_two[0])
            self.assertIs(model_two_1002, models_two[1])
            
            models_two = db_adapter.related_x_to_many_contains(
                model_one_1000, f_key, [model_two_1001, model_two_2002])
            self.assertEqual(len(models_two), 1)
            self.assertEqual(models_two[0].f_integer, 1001)
            self.assertIs(model_two_1001, models_two[0])
            
            models_two = db_adapter.related_x_to_many_contains(
                model_one_1000, f_key, [model_two_2001, model_two_2002])
            self.assertEqual(len(models_two), 0)
            
            models_two = db_adapter.related_x_to_many_contains(
                model_one_1000, f_key, [])
            self.assertEqual(len(models_two), 0)
    


from save_to_db.core.item import Item
from save_to_db.utils.test_base import TestBase



class TestItemClsManager(TestBase):
    """ Contains tests for
    :py:class:`~save_to_db.core.item_cls_manager.ItemClsManager`.
    """

    ModelGeneralOne = None
    ModelGeneralTwo = None


    def test_getting_items_by_path(self):
        #--- module `.m0.m1` ---
        m0_m1_module = '{}.m0.m1'.format(TestItemClsManager.__module__)
        
        M1_ItemA = type('M1_ItemA', (Item,), {})
        M1_ItemA.__module__ = m0_m1_module
                
        M1_ItemB = type('M1_ItemB', (Item,), {})
        M1_ItemB.__module__ = m0_m1_module
        
        M1_ItemRepeated = type('ItemRepeated', (Item,), {})
        M1_ItemRepeated.__module__ = m0_m1_module
        
        #--- module `.m0.m2` ---
        m0_m2_module = '{}.m0.m2'.format(TestItemClsManager.__module__)
        
        M2_ItemA = type('M2_ItemA', (Item,), {})
        M2_ItemA.__module__ = m0_m2_module
        
        M2_ItemB = type('M2_ItemB', (Item,), {})
        M2_ItemB.__module__ = m0_m2_module
        
        M2_ItemRepeated = type('ItemRepeated', (Item,), {})
        M2_ItemRepeated.__module__ = m0_m2_module
        
        # getting one item -----------------------------------------------------
        paths = (
            # absolute
            '{}.{}'.format(M1_ItemA.__module__, M1_ItemA.__name__),
            # class name only
            M1_ItemA.__name__,
            # relative one dot
            '.m0.m1.M1_ItemA',
            # relative multiple dots
            '..test_item_cls_manager.m0.m1.M1_ItemA',
        )
        this_module = TestItemClsManager.__module__
        
        for path in paths:
            if not path.startswith('.'):
                items = self.item_cls_manager.get_by_path(path)
            else:
                items = self.item_cls_manager.get_by_path(
                    path, relative_to=this_module)
            
            # although module can contain multiple classes with the same name,
            # in this case we know that there are only one such class
            self.assertTrue(len(items) == 1,
                            'One item must be returned: {}'.format(path))

            self.assertIn(M1_ItemA, items,
                          'Item was not found in registry, '
                          'path: {}'.format(path))
        
        # getting multiple items -----------------------------------------------
        path = 'ItemRepeated'
        items = self.item_cls_manager.get_by_path(path)
        self.assertGreaterEqual(len(items), 2,
                                'Not all items were returned by class name')
        
        self.assertTrue(len(items) == 2,
                        'Two items must be returned: {}'.format(path))
        
        self.assertIn(M1_ItemRepeated, items, 'Item was not found')
        self.assertIn(M2_ItemRepeated, items, 'Item was not found')
    
    
    def test_item_autogeneration(self):
        self.item_cls_manager.autogenerate = True
        
        # --- auto generated item two ---
        class ItemGeneralOne_A(Item):
            model_cls = self.ModelGeneralOne
        
        item_one = ItemGeneralOne_A()
        # item two class will be auto generated
        item_one['two_1_1'](f_boolean=True)
        
        ItemGeneralTwo11 = ItemGeneralOne_A.relations['two_1_1']['item_cls']
        ItemGeneralTwoXX = ItemGeneralOne_A.relations['two_x_x']['item_cls']
        
        self.assertIs(ItemGeneralTwo11, ItemGeneralTwoXX)
        
        self.assertFalse(ItemGeneralOne_A.metadata['autogenerated_item_cls'])
        self.assertTrue(ItemGeneralTwo11.metadata['autogenerated_item_cls'])
        self.assertTrue(ItemGeneralTwoXX.metadata['autogenerated_item_cls'])
        
        # --- using existing  item two ---
        self.item_cls_manager.clear()
        class ItemGeneralOne_B(Item):
            model_cls = self.ModelGeneralOne
            
        class ItemGeneralTwo_B(Item):
            model_cls = self.ModelGeneralTwo
        
        item_one = ItemGeneralOne_B()
        # item two class will be auto generated
        item_one['two_1_1'](f_boolean=True)
        
        ItemGeneralTwo11 = ItemGeneralOne_B.relations['two_1_1']['item_cls']
        ItemGeneralTwoXX = ItemGeneralOne_B.relations['two_x_x']['item_cls']
        
        self.assertIs(ItemGeneralTwo11, ItemGeneralTwo_B)
        self.assertIs(ItemGeneralTwoXX, ItemGeneralTwo_B)
        
        self.assertFalse(ItemGeneralOne_B.metadata['autogenerated_item_cls'])
        self.assertFalse(ItemGeneralTwo11.metadata['autogenerated_item_cls'])
        self.assertFalse(ItemGeneralTwoXX.metadata['autogenerated_item_cls'])


import sys

from .auto_reverse import *
from .constraints import *
from .fields import *
from .general import *
from .many_refs import *
from .merge import *


if 'manage.py' not in sys.argv and './manage.py' not in sys.argv:
    # Django will not allow to create migrations for the models
    from .invalid_fields import *


import sqlalchemy as sa

from ..config import Base



class ModelFieldTypes(Base):
    __tablename__ = 'model_field_types'
    
    id = sa.Column(sa.Integer, primary_key=True)
    
    binary_1 = sa.Column('new_binary_1', sa.Binary, nullable=True)
    binary_2 = sa.Column('new_binary_2', sa.BINARY, nullable=True)
    binary_3 = sa.Column('new_binary_3', sa.LargeBinary, nullable=True)
    binary_4 = sa.Column('new_binary_4', sa.VARBINARY, nullable=True)
    
    string_1 = sa.Column('new_string_1', sa.String, nullable=True)
    string_2 = sa.Column('new_string_2', sa.Unicode, nullable=True)
    string_3 = sa.Column('new_string_3', sa.VARCHAR, nullable=True)
    string_4 = sa.Column('new_string_4', sa.NVARCHAR, nullable=True)
    string_5 = sa.Column('new_string_5', sa.CHAR, nullable=True)
    string_6 = sa.Column('new_string_6', sa.NCHAR, nullable=True)
    
    text_1 = sa.Column('new_text_1', sa.Text, nullable=True)
    text_2 = sa.Column('new_text_2', sa.TEXT, nullable=True)
    text_3 = sa.Column('new_text_3', sa.UnicodeText, nullable=True)
    text_4 = sa.Column('new_text_4', sa.CLOB, nullable=True)
    
    integer_1 = sa.Column('new_integer_1', sa.Integer, nullable=True)
    integer_2 = sa.Column('new_integer_2', sa.INT, nullable=True)
    integer_3 = sa.Column('new_integer_3', sa.INTEGER, nullable=True)
    integer_4 = sa.Column('new_integer_4', sa.BigInteger, nullable=True)
    integer_5 = sa.Column('new_integer_5', sa.BIGINT, nullable=True)
    integer_6 = sa.Column('new_integer_6', sa.SmallInteger, nullable=True)
    integer_7 = sa.Column('new_integer_7', sa.SMALLINT, nullable=True)
    
    boolean_1 = sa.Column('new_boolean_1', sa.Boolean, nullable=True)
    boolean_2 = sa.Column('new_boolean_2', sa.BOOLEAN, nullable=True)
    
    float_1 = sa.Column('new_float_1', sa.Float(asdecimal=False),
                        nullable=True)
    float_2 = sa.Column('new_float_2', sa.FLOAT(asdecimal=False),
                        nullable=True)
    float_3 = sa.Column('new_float_3', sa.REAL(asdecimal=False),
                        nullable=True)
    float_4 = sa.Column('new_float_4', sa.DECIMAL(asdecimal=False),
                        nullable=True)
    float_5 = sa.Column('new_float_5', sa.Numeric(asdecimal=False),
                        nullable=True)
    float_6 = sa.Column('new_float_6', sa.NUMERIC(asdecimal=False),
                        nullable=True)
    
    decimal_1 = sa.Column('new_decimal_1', sa.Float(asdecimal=True),
                        nullable=True)
    decimal_2 = sa.Column('new_decimal_2', sa.FLOAT(asdecimal=True),
                        nullable=True)
    decimal_3 = sa.Column('new_decimal_3', sa.REAL(asdecimal=True),
                        nullable=True)
    decimal_4 = sa.Column('new_decimal_4', sa.DECIMAL(asdecimal=True),
                        nullable=True)
    decimal_5 = sa.Column('new_decimal_5', sa.Numeric(asdecimal=True),
                        nullable=True)
    decimal_6 = sa.Column('new_decimal_6', sa.NUMERIC(asdecimal=True),
                        nullable=True)
    
    date_1 = sa.Column('new_date_1', sa.Date, nullable=True)
    date_2 = sa.Column('new_date_2', sa.DATE, nullable=True)
    
    time_1 = sa.Column('new_time_1', sa.Time, nullable=True)
    time_2 = sa.Column('new_time_2', sa.TIME, nullable=True)
    
    datetime_1 = sa.Column('new_datetime_1', sa.DateTime, nullable=True)
    datetime_2 = sa.Column('new_datetime_2', sa.DATETIME, nullable=True)
    datetime_3 = sa.Column('new_datetime_3', sa.TIMESTAMP, nullable=True)
    





from save_to_db.adapters.sqlalchemy_adapter import SqlachemyAdapter
from save_to_db.core.persister import Persister

from . import config
from .sqlalchemy_test_base import SqlalchemyTestBase

from .models.fields import ModelFieldTypes
from .models.constraints import (ModelConstraintsOne,
                                 ModelConstraintsTwo,
                                 ModelConstraintsThree,
                                 ModelConstraintsFour,
                                 ModelConstraintsFive,
                                 ModelConstraintsSix,
                                 ModelConstraintsSelf)
from .models.general import (ModelGeneralOne,
                             ModelGeneralTwo)
from .models.invalid_fields import ModelInvalidFieldNames
from .models.many_refs import ModelManyRefsOne, ModelManyRefsTwo
from .models.auto_reverse import (ModelAutoReverseOne,
                                  ModelAutoReverseTwoA,
                                  ModelAutoReverseTwoB,
                                  ModelAutoReverseThreeA,
                                  ModelAutoReverseThreeB,
                                  ModelAutoReverseFourA,
                                  ModelAutoReverseFourB)
from .models.merge import (ModelMergeOne,
                           ModelMergeTwo,
                           ModelMergeThree,
                           ModelMergeOneX,
                           ModelMergeTwoX,
                           ModelMergeThreeX) 

from .. import register_test
from ..test_fields_and_relations import TestFieldsAndRelations
from ..test_constraints import TestConstraints
from ..test_item_cls_manager import TestItemClsManager
from ..test_merge_policy import TestMergeModels, TestPolicySetup
from ..test_item_persist import (TestAdapter,
                                 TestExceptions,
                                 TestItemHooks,
                                 TestResolveModel,
                                 TestModelDeleter,
                                 TestNorewrite,
                                 TestPersist,
                                 TestReplace,
                                 TestSimpleConfig)
from ..test_item_use import (TestBasicUse,
                             TestItemDump,
                             TestItemProcess,
                             TestItemSetup)
from ..test_scope import TestScope
from ..test_select import TestSelect
from ..test_signals import TestSignals

AdapterPrefix = 'Sqlalchemy'
persister = Persister(SqlachemyAdapter({'session': config.session,
                                        'ModelBase': config.Base}),
                      autocommit=True)
models = {
    'ModelFieldTypes': ModelFieldTypes,
    
    'ModelGeneralOne': ModelGeneralOne,
    'ModelGeneralTwo': ModelGeneralTwo,
    
    'ModelConstraintsOne': ModelConstraintsOne,
    'ModelConstraintsTwo': ModelConstraintsTwo,
    'ModelConstraintsThree': ModelConstraintsThree,
    'ModelConstraintsFour': ModelConstraintsFour,
    'ModelConstraintsFive': ModelConstraintsFive,
    'ModelConstraintsSix': ModelConstraintsSix,
    'ModelConstraintsSelf': ModelConstraintsSelf,
    
    'ModelInvalidFieldNames': ModelInvalidFieldNames,
    
    'ModelManyRefsOne': ModelManyRefsOne,
    'ModelManyRefsTwo': ModelManyRefsTwo,
        
    'ModelAutoReverseOne': ModelAutoReverseOne,
    'ModelAutoReverseTwoA': ModelAutoReverseTwoA,
    'ModelAutoReverseTwoB': ModelAutoReverseTwoB,
    'ModelAutoReverseThreeA': ModelAutoReverseThreeA,
    'ModelAutoReverseThreeB': ModelAutoReverseThreeB,
    'ModelAutoReverseFourA': ModelAutoReverseFourA,
    'ModelAutoReverseFourB': ModelAutoReverseFourB,
    
    'ModelMergeOne': ModelMergeOne,
    'ModelMergeTwo': ModelMergeTwo,
    'ModelMergeThree': ModelMergeThree,
    'ModelMergeOneX': ModelMergeOneX,
    'ModelMergeTwoX': ModelMergeTwoX,
    'ModelMergeThreeX': ModelMergeThreeX,
}

for test_cls in (TestFieldsAndRelations,
                 TestConstraints,
                 TestItemClsManager,
                 TestMergeModels,
                 TestPolicySetup,
                 TestAdapter,
                 TestExceptions,
                 TestItemHooks,
                 TestResolveModel,
                 TestModelDeleter,
                 TestNorewrite,
                 TestPersist,
                 TestReplace,
                 TestSimpleConfig,
                 TestBasicUse,
                 TestItemDump,
                 TestItemProcess,
                 TestItemSetup,
                 TestItemDump,
                 TestScope,
                 TestSelect,
                 TestSignals):

    register_test(AdapterPrefix, SqlalchemyTestBase,
                  test_cls, models, persister)




from django.db import models


class CommonColumns(models.Model):
    class Meta:
        abstract = True
        
    f_binary = models.BinaryField(null=True)
    f_boolean = models.NullBooleanField(null=True)
    f_string = models.CharField(max_length=255, null=True)
    f_text = models.TextField(null=True)
    f_integer = models.IntegerField(null=True)
    f_float = models.FloatField(null=True)
    f_date = models.DateField(null=True)
    f_time = models.TimeField(null=True)
    f_datetime = models.DateTimeField(null=True)


class ModelMergeOne(CommonColumns):
    class Meta:
        db_table = 'model_merge_one'
    
    two_1_1 = models.OneToOneField('ModelMergeTwo',
                                   on_delete=models.CASCADE,
                                   related_name='one_1_1',
                                   null=True)
    two_x_1 = models.ForeignKey('ModelMergeTwo',
                                on_delete=models.CASCADE,
                                related_name='one_1_x',
                                null=True)


class ModelMergeTwo(CommonColumns):
    class Meta:
        db_table = 'model_merge_two'
    
    three_1_1 = models.OneToOneField('ModelMergeThree',
                                   on_delete=models.CASCADE,
                                   related_name='two_1_1',
                                   null=True)
    three_x_1 = models.ForeignKey('ModelMergeThree',
                                  on_delete=models.CASCADE,
                                  related_name='two_1_x',
                                  null=True)


class ModelMergeThree(CommonColumns):
    class Meta:
        db_table = 'model_merge_three'
        

class ModelMergeOneX(CommonColumns):
    class Meta:
        db_table = 'model_merge_one_x'
        unique_together = (('f_integer', 'two_x_1'),)

    two_x_1 = models.ForeignKey('ModelMergeTwoX',
                                on_delete=models.CASCADE,
                                related_name='one_1_x',
                                null=True)


class ModelMergeTwoX(CommonColumns):
    class Meta:
        db_table = 'model_merge_two_x'
        unique_together = (('f_integer', 'three_x_1'),)

    three_x_1 = models.ForeignKey('ModelMergeThreeX',
                                  on_delete=models.CASCADE,
                                  related_name='two_1_x',
                                  null=True)


class ModelMergeThreeX(CommonColumns):
    class Meta:
        db_table = 'model_merge_three_x'
        


import re
from datetime import datetime
from decimal import Decimal
from pprint import pprint

from save_to_db.adapters.utils.column_type import ColumnType

from .item_base import ItemBase
from .item_contructor import complete_item_structure
from .item_metaclass import ItemMetaclass

from .bulk_item import BulkItem
from save_to_db.exceptions import MultipleModelsMatch, WrongAlias


#: regex for preparing strings to be converted into numeric values
_number_regex = re.compile(r'[^0-9\.]')


class Item(ItemBase, metaclass=ItemMetaclass, no_setup=True):
    """ This class is used to collect data for a single item in order to create
    or update corresponding row or rows in a database.
    
    **To configure an item class use next class variables:**
    
    :cvar model_cls: Reference to an ORM model class of one of the supported
        ORM libraries.
        
        .. note::
            You can use :py:func:`~save_to_db.info.print_info` function from
            :py:mod:`save_to_db.info` module to print to console all supported
            ORM library names and their configurations for use with this
            library.
    
    :cvar batch_size: Maximum number of models that can be pulled from database
        with one query.
        Default: `None` (defined by used ORM library itself).
        
            .. seealso::
                `BATCH_SIZE` class constant of
                :py:class:`~save_to_db.adapters.utils.adapter_base.AdapterBase`
                class.
    
    :cvar defaults: Dictionary with default values, where value can be
        either raw value or a callable that accepts an item as an argument
        and returns a field value.
        
        .. note::
            Defaults order is based on the number of '__' that keys contain,
            shorter keys are prioritized. This is due to the fact that default
            values can be another items.
        
        .. warning::
            When using other item instances as default values, do not forget
            about possible item scoping (see :py:class:`~.scope.Scope` class),
            make sure that related items assigned by default are from the same
            scope. You can use the fact that items are automatically created
            when accessed:
            
            .. code-block:: Python
                
                defaults = {
                    # to make default values compatible with scoping,
                    # use this:
                    'item_1': lambda item: item['some_key'](field='value')
                    # instead of this:
                    'item_2': lambda item: OtherItem(field='value')
                    # or this:
                    'item_3': OtherItem(field='value')
                }
        
    :cvar creators: List of groups of fields (sets) of an item. When values for
        all fields of a group are present in an item, a new record can be
        created in a database using that group.
    
    :cvar autoinject_creators: If `True` (default) and `creators_autoconfig` is
        also set to `True`, then not null model fields will be added to each
        group in *creators* list.
        Default: `True`.
    
    :cvar creators_autoconfig: Sets auto configuration for `creators` fields,
        possible values:
        
            - `True` - auto configuration is on (adds to `creators` if they were
              manually configured);
            - `False` - auto configuration is off;
            - `None` - auto configuration is on if `creators` was not
              configured manually, off otherwise (Default).
    
    :cvar getters: List of groups of fields (sets) of an item. When values for
        all fields of a group are present in an item, then the group can be
        used to look for a record in a database.
    
    :cvar getters_autoconfig: Same as `creators_autoconfig` but for `getters`
        fields.
    
    :cvar nullables: Set of field names for which value must be set to null or
        to an empty list (in case of x-to-many relationship) when saving, if
        value is not present in the item.
    
    :cvar remove_null_fields: Set of field names which, in case they have `None`
        value, must be removed from an item upon processing.
        
        .. warning::
            If a field name is listed in `nullables`, that it will not be
            deleted.
    
    :cvar remove_null_fields_autoconfig: Sets auto configuration for
        `remove_null_fields` fields, possible values:
        
            - `True` - auto configuration is on (adds to `remove_null_fields`
              if it was manually configured);
            - `False` - auto configuration is off;
            - `None` - auto configuration is on if `remove_nulls` was not
              configured manually, off otherwise (Default).
    
    :cvar relations:  dictionary describing foreign keys (relations),
        example:
            
            .. code-block:: Python
            
                relations = {
                    'one': {
                        'item_cls': ItemClassOne,
                        'replace_x_to_many': False,  # default value
                    }
                    'two': ItemClassTwo,  # if we only interested in item class
                }
        
        Keys are the fields that reference other items (foreign key columns
        in database) and values are dictionaries with next keys and values:
          
            - 'item_cls' - item class used to create related item;
            - 'replace_x_to_many' - Only applicable to x-to-many
              relationships. If `True` then when saving an item to a
              database, remove old values for the field from database.
              Default: `False`.
    
    :cvar aliases: A `dict` with field aliases. Dictionary keys are used as
        aliases, and values are used as actual item field names. Aliases and
        field names can contain double underscores ("__").
    
    :cvar conversions: A dictionary that describes how string values must be
        converted to proper data types. Default `conversions` value:
          
        .. code-block:: Python
          
            conversions = {
                'decimal_separator': '.',
                'boolean_true_strings': ('true', 'yes', 'on', '1', '+',),
                'boolean_false_strings': ('false', 'no', 'off', '0', '-',),
                # format values for dates and times used as arguments for
                # `datetime.datetime` function
                'date_formats': ('%Y-%m-%d',),  # can be multiple formats
                'time_formats': ('%H:%M:%S',),
                'datetime_formats': ('%Y-%m-%d %H:%M:%S',),
            }
        
        In case of absence of a value from `conversions` dictionary default
        value will be used.
        
        .. note::
            For "date_formats", "time_formats" and "datetime_formats" when
            only a single value is used, you can use that value instead of a
            tuple (or a list).
        
    :cvar allow_multi_update: If `True` then an instance of this class can
        update multiple models.
        Default: `False`.
    
    :cvar allow_merge_items:
        If `True` then all items that can potentially pull the same model from a
        database when persisting are merged into one item.
        Default: `False`.
    
    :cvar update_only_mode: If `True` then a new model in a database will not be
        created if it does not exist.
        Default: `False`.
        
        .. note:: Can be overwritten on an instance.
    
    :cvar get_only_mode: If `True` then the item data is used only to pull
        models from database in order to load other models through
        relationships.
        Default: `False`.
        
        .. note:: Can be overwritten on an instance.
        
    :cvar norewrite_fields: Dictionary with fields that cannot be changed.
        Example:
    
        .. code-block:: Python
        
            from save_to_db import RelationType
        
            norewrite_fields = {
                # always can rewrite
                'field_1': None,
                # rewrite value if it equals to`None`
                'field_2': True,
                # never rewrite, set only when model is created
                'field_3': False,
                # you can use tuples as keys
                ('field_4', 'field_5'): None,
                # referring to all many-to-many fields
                # (relation type can be used as a key)
                RelationType.MANY_TO_MANY: None,
                True: False,  # `True` for all the other fields
            }
        
        .. note::
            
            - You can use a `tuple` as a key to set many fields at once.
            - You can use `True` key for all not listed fields, including
              relations.
            - You can use values from
              :py:class:`~save_to_db.adapters.utils.relation_type.RelationType`
              enumeration for relation fields.
        
        .. note::
            If norewrite value is `True` for a field (can rewrite `None`),
            for many-to-x relationships the "many" side, even if already
            existed, can be added to by newly created related models. 
    
    :cvar fast_insert: If `True` then models will not be pulled from database
        when persisting to database, new models will be created without trying
        to update.
        Default: `False`.
    
    :cvar deleter_selectors: Set of item field names that are going to
        be collected from processed (created and updated) ORM models and
        used as selectors.
        See `deleter_keepers` for more details.
    
    :cvar deleter_keepers: Set of item field names that are going to
        be collected from processed (created and updated) ORM models and used
        as keepers.
        Default value: `None` if `deleter_selectors` are also `None`,
        set of primary key fields otherwise.
        
        After finishing working with all items, you can call
        :py:meth:`~.persister.Persister.execute_deleter` method or
        :py:meth:`~.persister.Persister.execute_scope_deleter` method of
        :py:class:`~.persister.Persister` class to delete all those models that
        have same field values that were collected using
        `deleter_selectors`, but not the same as collected using
        `deleter_keepers`.
        
        .. seealso::
            `deleter_execute_on_persist` setting.
        
        .. seealso::
            :py:class:`~save_to_db.core.model_deleter.ModelDeleter` class.
            An instance of the class is created and stored in
            `item_cls.metadata['model_deleter']` class variable (where
            `item_cls` is a subclass of this class).
    
    :cvar deleter_execute_on_persist:
        If `True` then model deleter is executed upon item persistence.
        Default: `False`.
    
    :cvar unref_x_to_many: A dictionary containing x-to-many relationship
        fields as keys, and model deleter settings as values.
        
        Example:
        
        .. code-block:: Python
        
            unref_x_to_many = {
                'field_1_x_first': {
                    'selectors': ['field_1', 'field_2'],
                    # default for keepers: set of primary key fields
                    'keepers': ['field_a', 'field_b'],
                },
                # if you want to use default keepers, you can use a shortcut
                'field_1_x_first': ['field_x', 'field_y']
            }
        
        Upon saving referenced by the fields models are unreferenced (removed
        from relation) if they can be selected from database by selectors,
        but cannot be selected using keepers.
        
        Selectors and keepers values are grabbed from the items being added to
        relations.
        
        .. seealso::
            `deleter_selectors` and `deleter_keepers`.
            
        .. seealso::
            :py:class:`~save_to_db.core.model_deleter.ModelDeleter` class.
            Instances of the class are created and stored in
            `item_cls.metadata['model_unrefs']` class dictionary (where
            `item_cls` is a subclass of this class). Dictionary keys are
            x-to-many field names, dictionary values are instances of
            :py:class:`~save_to_db.core.model_deleter.ModelDeleter` class.
    """
    
    DUMP_INSTANCE_ATTRS = ('update_only_mode', 'get_only_mode')
    
    
    #--- special methods -------------------------------------------------------

    @classmethod
    def _get_real_keys(cls, key, as_string=False):
        result = []
        item_cls = cls

        def add_key_to_result(add_key):
            nonlocal result, item_cls

            item_cls.complete_setup()
            
            if add_key in item_cls.relations:
                item_cls = item_cls.relations[add_key]['item_cls']
                result.append(add_key)
            elif add_key in item_cls.fields:
                item_cls = None
                result.append(add_key)
            else:
                raise WrongAlias('Wrong key: {} (full path: {}), '
                                 'alias of: {}'.format(add_key, key, cls))
            
        cur_key = key
        while cur_key:
            if item_cls is None:
                raise WrongAlias(
                    'Cannot process the rest of path: {} (full path: {}), '
                    'alias of: {}'.format(cur_key, key, cls))
            
            aliased = False
            for alias, path in item_cls.aliases.items():
                if cur_key.startswith(alias):
                    cur_key = cur_key.replace(alias, path, 1)
                    aliased = True
                    break
            if aliased:
                continue
            
            if '__' in cur_key:
                add_key, cur_key = cur_key.split('__', 1)
                add_key_to_result(add_key)
            else:
                add_key_to_result(cur_key)
                break

        if not as_string:
            return result
        return '__'.join(result)
    
    
    @classmethod
    def _genitem(cls, key):
        """ Generates an item for the given relation key.
        
        :param key: Relation key for which to create related item instance.
        :returns: Generated item.
        """
        relation = cls.relations[key]
        if relation['relation_type'].is_x_to_many():
            item = relation['item_cls'].Bulk()
        else:
            item = relation['item_cls']()
        return item
    
    
    def __setitem__(self, key, value):
        item = self
        real_keys = item._get_real_keys(key)
        for i, real_key in enumerate(real_keys[:-1]):
            if real_key not in item.data:
                item.data[real_key] = item._genitem(real_key)
    
            item = item.data[real_key]
            if item.is_bulk_item():
                item.data['__'.join(real_keys[i+1:])] = value
                return
        
        item.data[real_keys[-1]] = value
        
        
    def __getitem__(self, key):
        result = self
        real_keys = result._get_real_keys(key)
        for i, real_key in enumerate(real_keys):
            result.complete_setup()
            
            if real_key not in result.data:
                if real_key in result.fields:
                    raise KeyError(real_key)
                
                new_item = result._genitem(real_key)
                result.data[real_key] = new_item
                
                # setting reverse key
                relation = result.relations[real_key]
                reverse_key = relation['reverse_key']
                   
                if reverse_key:
                    if new_item.is_single_item():
                        if relation['relation_type'].is_one_to_x():
                            new_item[reverse_key] = result
                        else:
                            new_item[reverse_key].add(result)
                    else:
                        # if it is a bulk item, just making it a default value
                        if relation['relation_type'].is_one_to_x():
                            new_item[reverse_key] = result
                        else:
                            result_as_bulk = result.Bulk()
                            result_as_bulk.add(result)
                            new_item[reverse_key] = result_as_bulk
                
            result = result.data[real_key]
            if isinstance(result, BulkItem):
                bulk_real_keys = real_keys[i+1:]
                if not bulk_real_keys:
                    # return the bulk item itself
                    return result
                # something in bulk item
                return result._get_direct(bulk_real_keys)
        
        return result
    
    
    def __delitem__(self, key):
        real_keys = self._get_real_keys(key)
        item = self
        for i, real_key in enumerate(real_keys[:-1]):
            item = item.data[real_key]
            if isinstance(item, BulkItem):
                real_key = '__'.join(real_keys[i+1:])
                del item.data[real_key]
                return
            
        del item.data[real_keys[-1]]
    
    
    def __contains__(self, key):
        try:
            real_keys = self._get_real_keys(key)
        except KeyError:
            return False
        return self._contains_direct(real_keys)
    
    
    def _contains_direct(self, real_keys):
        this_key, *those_keys = real_keys
        if this_key in self.data:
            if not those_keys:
                return True
            return self.data[this_key]._contains_direct(those_keys)
        return False
    
    #--- utility methods -------------------------------------------------------
    
    def to_dict(self, revert=False):
        return self._to_dict(revert=revert,
                             _item_to_dict={}, _address_to_item={})
    
    
    def _to_dict(self, revert, _item_to_dict, _address_to_item):
        self_address = id(self)
        if self_address in _item_to_dict:
            if self_address not in _address_to_item:
                _address_to_item[self_address] = len(_address_to_item) + 1
            self_id = _address_to_item[self_address]
            _item_to_dict[self_address]['id'] = self_id
            return {
                'id': self_id,
            }
        
        result = {
            'item': {},
        }
        _item_to_dict[self_address] = result
        
        # relations have to be always processed in the same order to have
        # the same value under "id" key in result dictionary each time
        data_keys = list(self.data.keys())
        data_keys.sort()
        for key in data_keys:
            value = self.data[key]
            if not isinstance(value, ItemBase):
                if revert:
                    value = self.revert_field(key, value, aliased=False)
                result['item'][key] = value
            else:
                result['item'][key] = \
                    value._to_dict(revert=revert,
                                   _item_to_dict=_item_to_dict,
                                   _address_to_item=_address_to_item)
        
        cls = type(self)
        for key in self.DUMP_INSTANCE_ATTRS:
            self_value = getattr(self, key)
            cls_value = getattr(cls, key)
            if self_value != cls_value:
                result[key] = self_value
        
        return result
    
    
    def load_dict(self, data):
        return self._load_dict(data, _id_to_item={})
    
    
    def _load_dict(self, data, _id_to_item):
        
        if 'id' in data:
            if data['id'] not in _id_to_item:
                _id_to_item[data['id']] = type(self)()
            
            item = _id_to_item[data['id']]
            if 'item' not in data:
                return item
        else:
            item = type(self)()
        
        if 'item' not in data:
            raise Exception(data.keys())
        
        for key, value in data['item'].items():
            if key in item.fields:
                item[key] = value
            elif key in item.relations:
                item[key] = item[key]._load_dict(value, _id_to_item)
        
        for key in self.DUMP_INSTANCE_ATTRS:
            if key in data:
                setattr(item, key, data[key])
                
        return item
    
    
    @classmethod
    def get_item_cls(cls):
        return cls
    
    @classmethod
    def is_scoped(cls):
        return cls.metadata['collection_id'] != None
    
    @classmethod
    def get_collection_id(cls):
        return cls.metadata['collection_id']
    
    
    #--- main methods ----------------------------------------------------------
    

    @staticmethod
    def _Auto(model_cls, **kwargs):
        """ Creates new item class.
        
        :param model_cls: model for witch new item class is generated.
        :param \*\*kwargs: key-word arguments that will be used as new class
            attributes.
        :returns: Newly created item class.
        """
        kwargs['model_cls'] = model_cls
        item_cls = type('{}Item'.format(model_cls.__name__),
                        (Item,), kwargs)
        return item_cls
        
    
    @classmethod
    def Bulk(cls, *args, **kwargs):
        """ Creates a :py:class:`~.bulk_item.BulkItem` instance for this item
        class.
        
        :param \*args: Positional arguments that are passed to bulk item
            constructor.
        :param \*\*kwargs: Keyword arguments that are passed to bulk item
            constructor.
        :returns: :py:class:`~.bulk_item.BulkItem` instance for this item class.
        """
        return BulkItem(cls, *args, **kwargs)
    
    
    def as_list(self):
        return [self]
    
    
    def is_single_item(self):
        return True
    
    
    def is_bulk_item(self):
        return False
    
    
    @classmethod
    def revert_field(cls, key, value, aliased=True):
        """ Converts field into JSON serializable field in such a way
        that :py:meth:`~.process_field` method converts it back to the original
        value.
        
        :param key: Key using which proper value type can be determined.
            This value can contain double underscores to reference relations.
        :param value: Value to process.
        :param aliased: If it's `True` then `key` contains field aliases.
        :returns: Value converted to proper type.
        """
        if isinstance(value, str):
            return value
        
        if value is None:
            return
        
        if aliased:
            real_key = cls._get_real_keys(key, as_string=True)
        else:
            real_key = key
            
        if '__' in real_key:
            this_key, that_key = real_key.split('__', 1)
            return cls.relations[this_key]['item_cls'].revert_field(
                that_key, value, aliased=False)
        
        conversions = cls.conversions
        column_type = cls.fields[real_key]
        
        if column_type is ColumnType.DECIMAL:
            return str(value).replace('.', conversions['decimal_separator'])

        elif column_type is ColumnType.DATE:
            return value.strftime(conversions['date_formats'][0])
        elif column_type is ColumnType.TIME:
            return value.strftime(conversions['time_formats'][0])
        elif column_type is ColumnType.DATETIME:
            return value.strftime(conversions['datetime_formats'][0])
        
        return value
                
    
    def revert(self):
        return self._revert(_procesed_items=[])
    
    
    def _revert(self, _procesed_items):
        if self in _procesed_items:
            return
        _procesed_items.append(self)
        
        for key, value in self.data.items():
            if not isinstance(value, ItemBase):
                try:
                    self.data[key] = self.revert_field(key, value,
                                                       aliased=False)
                except:
                    raise Exception(key, value, isinstance(value, ItemBase))
            else:
                value._revert(_procesed_items=_procesed_items)
    
    
    @classmethod
    def process_field(cls, key, value, aliased=True):
        """ Converts `value` to the appropriate data type for the given
        `key`.
        
        :param key: Key using which proper value type can be determined.
            This value can contain double underscores to reference relations.
        :param value: Value to process.
        :param aliased: If it's `True` then `key` contains field aliases.
        :returns: Value converted to proper type.
        """
        cls.complete_setup()
        
        if value is None:
            return None
        
        if aliased:
            real_key = cls._get_real_keys(key, as_string=True)
        else:
            real_key = key
        
        if '__' in real_key:
            this_key, that_key = real_key.split('__', 1)
            return cls.relations[this_key]['item_cls'].process_field(
                that_key, value, aliased=False)
        
        conversions = cls.conversions
        column_type = cls.fields[real_key]
        if column_type is ColumnType.BINARY:
            if not isinstance(value, bytes):
                if isinstance(value, str):
                    value = bytes(value, 'utf-8')
                else:
                    raise ValueError('Cannot convert to bytes: {}'.format(
                        value))
        elif column_type is ColumnType.BOOLEAN:
            if not isinstance(value, bool):
                if value.lower() in conversions['boolean_true_strings']:
                    value = True
                elif value.lower() in conversions['boolean_false_strings']:
                    value = False
                else:
                    raise ValueError('Cannot convert to boolean: {}'.format(
                        value))
        elif column_type.is_str():
            if not isinstance(value, str):
                value = str(value)
        elif column_type.is_num():
            if isinstance(value, str):
                if conversions['decimal_separator'] != '.':
                    value = value.replace('.', '')
                value = value.replace(conversions['decimal_separator'], '.')
                value = _number_regex.sub('', value)
            if column_type is ColumnType.INTEGER:
                if not isinstance(value, int):
                    value = int(value)
            elif column_type is ColumnType.FLOAT:
                if not isinstance(value, float):
                    value = float(value)
            elif column_type is ColumnType.DECIMAL:
                if not isinstance(value, Decimal):
                    value = Decimal(value)
        elif column_type in (ColumnType.DATE,
                             ColumnType.TIME,
                             ColumnType.DATETIME):
            
            if isinstance(value, str):
                format_list = {
                    ColumnType.DATE: conversions['date_formats'],
                    ColumnType.TIME: conversions['time_formats'],
                    ColumnType.DATETIME: conversions['datetime_formats'],
                }[column_type]

                for value_format in format_list:
                    try:
                        value = datetime.strptime(value, value_format)
                        break
                    except ValueError:
                        pass
                else:
                    raise Exception(
                        'Failed to convert key "{}", value: {}'.format(
                            key, value))
                
                # converting datetime if needed
                if column_type is ColumnType.DATE:
                    value = value.date()
                
                elif column_type is ColumnType.TIME:
                    value = value.time()

        return value
    
    
    def process(self):
        self._process(_procesed_items=[])
        return complete_item_structure(self)
            
            
    def _process(self, _procesed_items):
        if self in _procesed_items:
            return
        _procesed_items.append(self)
        
        self.before_process()  # hook
        
        # processing set fields
        for key, value in list(self.data.items()):  # changes size
            if not isinstance(value, ItemBase):
                try:
                    self.data[key] = self.process_field(key, value,
                                                        aliased=False)
                except:
                    raise Exception(key, value, isinstance(value, ItemBase))
            else:
                value._process(_procesed_items=_procesed_items)
        
        # processing defaults
        keys = list(self.defaults.keys())
        # 1. first processing shortest keys
        # 2. defaults that use functions must be processed at the end
        def get_key_for_default(key):
            value = self.defaults[key]
            order = 0
            if hasattr(value, '__call__') and not isinstance(value, ItemBase):
                order = 1
            # (although right now referencing related items is impossible here)
            return (key.count('__'), order)
        keys.sort(key=get_key_for_default)
        
        for key in keys:
            if key in self.data:
                continue
            
            value = self.defaults[key]
            if hasattr(value, '__call__') and not isinstance(value, ItemBase):
                value = value(self)
                
            end_item_cls = type(self)
            end_relation = None
            for subkey in key.split('__'):
                if subkey in end_item_cls.relations:
                    end_relation = end_item_cls.relations[subkey]
                    end_item_cls = end_relation['item_cls']
                else:
                    end_relation = None
            
            if not end_relation:  # not a relations
                self[key] = self.process_field(key, value, aliased=False)
            else:
                if not isinstance(value, list) or \
                        not end_relation['relation_type'].is_x_to_many():
                    value._process(_procesed_items=_procesed_items)
                    self[key] = value
                else:
                    bulk_item = self.relations[key]['item_cls'].Bulk()
                    bulk_item.add(*value)
                    bulk_item._process(_procesed_items=_procesed_items)
                    self[key] = bulk_item
        
        # processing nullables
        for key in self.nullables:
            if key in self:
                continue
            if key in self.fields:
                self[key] = None
            elif key in self.relations:
                relation = self.relations[key]
                if relation['relation_type'].is_x_to_many():
                    self[key] = relation['item_cls'].Bulk()
                    self[key]._process(_procesed_items=_procesed_items)
                else:
                    self[key] = None
        
        # processing remove_null_fields
        for key in self.remove_null_fields:
            if key not in self or key in self.nullables:
                continue
            if key not in self.relations or \
                    not self.relations[key]['relation_type'].is_x_to_many():
                if self[key] is None:
                    del self[key]
            else:
                if len(self[key]) == 0:
                    del self[key]
        
        self.after_process()  # hook
    
    
    def resolve_model(self, models):
        """ This function is called during persisting when two or more models
        match the item and `allow_multi_update` setting is `False`
        it must return a single item to be used or raise a
        :py:class:`save_to_db.exceptions.item_persist.MultipleModelsMatch`
        exception.
        
                
        .. warning::
            If `allow_multi_update` option is `True`, then this function is
            ignored.
        
        :param models: List of models that this item matches.
        :returns: Single model to be updated.
        """
        raise MultipleModelsMatch(self, models)
    
    
    #--- helper functions ------------------------------------------------------
    
    def get_config(self):
        """ Returns instance configuration. """
        
        default_configuration = ItemMetaclass.get_default_configuration()
        config = {}
        for key in default_configuration.keys():
            config[key] = getattr(self, key)
        return config
        
    
    def print_config(self):
        """ Pretty prints item instance configuration to console. """
        
        pprint(self.get_config())
    
    
    @classmethod
    def get_cls_config(cls):
        """ Returns class configuration. """
        
        return cls.get_config(cls)
    
    
    @classmethod
    def print_cls_config(cls):
        """ Pretty prints item class configuration to console. """
        
        pprint(cls.get_cls_config())
        
        
    #--- hooks -----------------------------------------------------------------
    
    
    def before_process(self):
        """ A hook method that is called before processing fields values. """
        pass
    
    
    def after_process(self):
        """ A hook method that is called immediately after all fields have been
        processed.
        """
        pass
    
    
    def before_model_update(self, model):
        """ A hook method that is called before updating matching model with
        item data.
        
        :param model: Model that was pulled from database or freshly
            created (in case there were no matching models).
        """
        pass
    
    
    def after_model_save(self, model):
        """ A hook method that is called after updating matching ORM model with
        item data and saving the model to a database.
        
        :param model: Model that was updated.
        """
        pass


from sqlalchemy import and_, or_, inspect
from sqlalchemy.schema import UniqueConstraint, PrimaryKeyConstraint
from sqlalchemy.sql import sqltypes
from sqlalchemy.ext.declarative.api import DeclarativeMeta

from .utils.adapter_base import AdapterBase
from .utils.column_type import ColumnType
from .utils.relation_type import RelationType



class SqlachemyAdapter(AdapterBase):
    """ An adapter for working with SqlAlchemy library.
    
    The `adapter_settings` for this class is a simple Python dictionary that
    must contain next values:
        
        - *session* is an SqlAlchemy session object that will be used to query
          a database.
        - *ModelBase* is a declarative base class for ORM models.
    """
    
    def __init__(self, adapter_settings):
        super().__init__(adapter_settings)
        self.session = adapter_settings['session']
        self.ModelBase = adapter_settings['ModelBase']
        
    
    #--- general methods -------------------------------------------------------
    
    @classmethod
    def is_usable(cls, model_cls):
        return issubclass(type(model_cls), DeclarativeMeta)
    
    
    def commit(self):
        self.session.commit()
    
    
    def rollback(self):
        self.session.rollback()
        
    
    @classmethod
    def iter_fields(cls, model_cls):
        for field in inspect(model_cls).attrs:
            if hasattr(field, 'target'):
                continue  # a relation
            
            column = list(field.columns)[0]
            if column.foreign_keys:
                continue  # foreign key
            
            column_type = type(column.type)
            yield_type = ColumnType.OTHER
            
            for sql_type, cur_yield_type in (
                    (sqltypes._Binary, ColumnType.BINARY),
                    # Text is derived from String class, so checking it firsts
                    (sqltypes.Text, ColumnType.TEXT),
                    (sqltypes.String, ColumnType.STRING),
                    (sqltypes.Integer, ColumnType.INTEGER),
                    (sqltypes.Numeric, ColumnType.FLOAT),  # can be cecimal
                    (sqltypes.Date, ColumnType.DATE),
                    (sqltypes.Time, ColumnType.TIME),
                    (sqltypes.DateTime, ColumnType.DATETIME),
                    (sqltypes.Boolean, ColumnType.BOOLEAN)):
                if issubclass(column_type, sql_type):
                    if cur_yield_type is ColumnType.FLOAT and \
                            column.type.asdecimal:
                        cur_yield_type = ColumnType.DECIMAL
                    yield_type = cur_yield_type
                    break
            
            yield field.key, yield_type
    
    
    @classmethod
    def iter_relations(cls, model_cls):
        registry = model_cls._decl_class_registry

        for field in inspect(model_cls).attrs:
            if not hasattr(field, 'target'):
                continue  # not a relation
            
            # getting referenced model
            other_model_cls = None
            for value_from_registry in registry.values():
                if hasattr(value_from_registry, '__table__') and \
                        value_from_registry.__table__ is field.target:
                    other_model_cls = value_from_registry
                    break
            
            # getting relation type
            direction_name = field.direction.name
            
            if direction_name == 'MANYTOMANY':
                direction = RelationType.MANY_TO_MANY
                
            elif direction_name == 'MANYTOONE':
                direction = RelationType.MANY_TO_ONE
                # maybe one to one
                other_field = field.back_populates or field.backref
                if other_field is not None:
                    other_attr = getattr(other_model_cls, other_field)
                    if not other_attr.property.uselist:
                        direction = RelationType.ONE_TO_ONE
                        
            elif direction_name == 'ONETOMANY':
                if field.uselist:
                    direction = RelationType.ONE_TO_MANY
                else:
                    direction = RelationType.ONE_TO_ONE
            
            # getting reverse key
            reverse_key = None
            if field.back_populates:
                reverse_key = field.back_populates
            elif field.backref:
                reverse_key = field.backref[0]
            
            yield (field.key, other_model_cls, direction, reverse_key)
    
    
    @staticmethod
    def __iter_fields_columns(model_cls):
        """ Returns an iterator of lists of type:
        
            [field_name, column_names]
        
        Where:
            - *field_name* is a name of a field of the `model_cls`.
            - *column_names* is a set of column names that the field uses
              (for example, in case of composite foreign key there will be
              one field that uses many columns to reference another model)
              
        :param model_cls: ORM model class for which relations are going to be
            iterated.
        :returns: A generator of field names and their columns.
        """
        
        for field in inspect(model_cls).attrs:

            if hasattr(field, 'columns') and \
                    list(field.columns)[0].foreign_keys:
                continue  # foreign key (not a relation)
            
            if hasattr(field, 'target'):  # a relation
                # assume that composite key has columns only from one table
                if not list(field.local_columns)[0].foreign_keys:
                    # no columns for the foreign key on this side
                    continue
                
                columns = field.local_columns
            else:
                columns = field.columns
            
            yield [field.key, {column.name for column in columns}]
    
    
    @classmethod
    def iter_required_fields(cls, model_cls):
        # --- collecting not null column names ---
        ignore_primary_key = False
        if len(model_cls.__table__.primary_key.columns) == 1:
            primary_column = list(model_cls.__table__.primary_key.columns)[0]
            if isinstance(primary_column.type, sqltypes.Integer):
                # single primary key of integer type have their default sequence
                # generated 
                ignore_primary_key = True
        
        required_column_names = set()
        for column_name, column in model_cls.__table__.columns.items():
            if hasattr(column.default, 'arg'):
                continue  # has a default value
            if not column.nullable:
                if not column.primary_key or not ignore_primary_key:
                    required_column_names.add(column_name)
                    continue
        
        # --- yielding field names ---
        for field_name, column_names in cls.__iter_fields_columns(model_cls):
            if required_column_names.intersection(column_names):
                yield field_name

    
    @classmethod
    def iter_unique_field_combinations(cls, model_cls):
        yielded_one_to_one = set()
        # --- one to one is also unique ---
        registry = model_cls._decl_class_registry

        for field in inspect(model_cls).attrs:
            if not hasattr(field, 'target'):
                continue  # not a relation
            
            direction_name = field.direction.name
            
            # getting referenced model
            other_model_cls = None
            for value_from_registry in registry.values():
                if hasattr(value_from_registry, '__table__') and \
                        value_from_registry.__table__ is field.target:
                    other_model_cls = value_from_registry
                    break
            
            if direction_name == 'MANYTOONE':
                other_field = field.back_populates or field.backref
                if other_field is not None:
                    other_attr = getattr(other_model_cls, other_field)
                    if not other_attr.property.uselist:  # one to one
                        yielded_one_to_one.add(field.key)
                        yield {field.key}
                        
            elif direction_name == 'ONETOMANY':
                if not field.uselist:  # one-to-one
                    yielded_one_to_one.add(field.key)
                    yield {field.key}
        
        # --- collecting unique constraints (for columns) ---
        unique_column_combinations = []
        for constraint in model_cls.__table__.constraints:
            if not isinstance(constraint, (UniqueConstraint,
                                           PrimaryKeyConstraint)):
                continue
            column_names = set()
            for column in constraint.columns:
                column_names.add(column.name)
                
            unique_column_combinations.append(column_names)
            
        # --- yielding unique field name combinations ---
        for unique_column_combination in unique_column_combinations:
            unique_fields = set()
            
            for field_name, column_names in \
                    cls.__iter_fields_columns(model_cls):
                if not (column_names - unique_column_combination):
                    unique_fields.add(field_name)
            if unique_fields:
                if len(unique_fields) == 1 and \
                        not (unique_fields - yielded_one_to_one):
                    continue  # was already yielded
                    
                yield unique_fields
        
        
    @classmethod
    def get_table_fullname(cls, model_cls):
        return model_cls.__table__.fullname
    
    
    def get_model_cls_by_table_fullname(self, name):
        for value_from_registry in self.ModelBase._decl_class_registry.values():
            if not hasattr(value_from_registry, '__table__'):
                continue
            if name ==  self.get_table_fullname(value_from_registry):
                return value_from_registry
    
    
    def iter_all_models(self):
        for value_from_registry in self.ModelBase._decl_class_registry.values():
            if hasattr(value_from_registry, '__table__'):
                yield value_from_registry
    

    #--- methods for working with items ----------------------------------------
    
    @staticmethod
    def __none_on_the_other_side(model_field, related_model):
        # if it is the other model that has reference to
        # this model with foreign key equals `None`, then we
        # know that other model is not usable as filter
        for column in model_field.property.remote_side:
            if getattr(related_model, column.name) is None:
                return True
        return False
    
    
    def get(self, items_and_fkeys):
        if not items_and_fkeys:
            return []
        
        # initializing variables -----------------------------------------------
        item_cls = items_and_fkeys[0][0]
        model_cls = item_cls.model_cls
        session = self.session
        
        #--- getting item models from database ---------------------------------
        all_items_filters = []
        query_object = session.query(model_cls)
        getters = items_and_fkeys[0][0].getters
        
        for item, fkeys in items_and_fkeys:
            one_item_filters = []
            for group in getters:
                group_filters = []
                # in case related item was not in database
                skip_group_filters = False
                for field_name in group:
                    if field_name not in item:
                        skip_group_filters = True
                        break
                     
                    model_field = getattr(item.model_cls,
                                          field_name)
                    if field_name in item.fields:
                        field_value = item[field_name]
                        group_filters.append(model_field==field_value)
                    elif field_name in item.relations:
                        related_models = fkeys.get(field_name)
                        if not related_models:
                            # failed to get or created related model before
                            skip_group_filters = True
                            break
                         
                        relation = item.relations[field_name]
                        if relation['relation_type'].is_x_to_one():
                            if related_models[0] is not None:
                                if self.__none_on_the_other_side(
                                        model_field, related_models[0]):
                                    skip_group_filters = True
                                    break
                         
                            group_filters.append(
                                model_field==related_models[0])
                        else:
                            contains_any = []
                            relation_type = \
                                item.relations[field_name]['relation_type']
                            is_x_to_x = \
                                relation_type is RelationType.MANY_TO_MANY
                            for related_model in related_models:
                                if is_x_to_x or \
                                        not self.__none_on_the_other_side(
                                            model_field, related_model):
                                    contains_any.append(
                                        model_field.contains(related_model))
                            if contains_any:
                                group_filters.append(or_(*contains_any))
                            else:
                                skip_group_filters = True
                                break
                     
                if not skip_group_filters and group_filters:
                    one_item_filters.append(and_(*group_filters))
             
            if one_item_filters:
                all_items_filters.append(or_(*one_item_filters))
        
        if not all_items_filters:
            return []  # nothing to get from DB
            
        query_object = query_object.filter(or_(*all_items_filters))
        models = query_object.all()
         
#         print('+'*40)
#         print(str(query_object.statement.compile(
#             compile_kwargs={"literal_binds": True})), flush=True)
            
        return models
    

    def delete(self, model):
        self.session.delete(model)
    

    def create_blank_model(self, model_cls):
        model = model_cls()
        self.session.add(model)
        return model
    

    def add_related_models(self, model, fkey, related_models):
        related_list = getattr(model, fkey)
        for model in related_models:
            if model not in related_list:
                related_list.append(model)
    

    def clear_related_models(self, model, fkey):
        getattr(model, fkey).clear()
        

    def related_x_to_many_exists(self, model, fkey):
        model_cls = type(model)
        exists_query = self.session.query(model_cls).filter(
            getattr(model_cls, fkey).any(),
            *[
                getattr(model_cls, pk) == getattr(model, pk)
                for pk in self.get_primary_key_names(model_cls)
            ]
        ).exists()
        return self.session.query(exists_query).scalar()
        

    def related_x_to_many_contains(self, model, fkey, child_models):
        if not child_models:
            return []
        
        def model_filters(models):
            model_cls = type(models[0])
            filters = []
            for model in models:
                filters.append(and_(*[
                    getattr(model_cls, pk) == getattr(model, pk)
                    for pk in self.get_primary_key_names(model_cls)
                ]))
            return or_(*filters)
        
        model_cls = type(model)
        child_model_cls = type(child_models[0])
        filter_value = and_(
            model_filters([model]),
            model_filters(child_models)
        )
        
        query_object = self.session.query(child_model_cls) \
            .join(getattr(model_cls, fkey)) \
            .filter(filter_value)
            
        return query_object.all()
    
    @classmethod
    def get_primary_key_names(cls, model_cls):
        return tuple(pk.name for pk in inspect(model_cls).primary_key)
    

    def save_model(self, model):
        self.session.add(model)
        self.session.flush()
    
    @classmethod
    def __get_select_and_keep_fielter(cls, model_cls, selectors, keepers):
        select_filter_sets = []
        for selectors_entry in selectors:
            select_filter_set = []
            for field_name, value in selectors_entry.items():
                select_filter_set.append(
                    getattr(model_cls, field_name) == value
                )
            select_filter_sets.append(and_(*select_filter_set))
        total_select_filter = or_(*select_filter_sets)
        
        keep_fielter_sets = []
        for keepers_entry in keepers:
            keep_fielter_set = []
            for field_name, value in keepers_entry.items():
                keep_fielter_set.append(
                    getattr(model_cls, field_name) != value
                )
            keep_fielter_sets.append(or_(*keep_fielter_set))
        total_keep_filter = and_(*keep_fielter_sets)
        
        if select_filter_sets or keep_fielter_sets:
            return and_(total_select_filter, total_keep_filter)
        return None
    

    def execute_delete(self, model_cls, selectors, keepers):  
        filters = self.__get_select_and_keep_fielter(model_cls,
                                                     selectors, keepers)
        if filters is not None:
            self.session.query(model_cls).filter(filters).delete()
    
    
    def execute_unref(self, parent, fkey, selectors, keepers):
        parent_cls = type(parent)
        registry = parent_cls._decl_class_registry
        target_property = getattr(parent_cls, fkey).property.target
        
        # getting referenced model class
        for value_from_registry in registry.values():
            if hasattr(value_from_registry, '__table__') and \
                    value_from_registry.__table__ is target_property:
                target_cls = value_from_registry
                break
        
        target_filters = self.__get_select_and_keep_fielter(target_cls,
                                                            selectors,
                                                            keepers)
        if target_filters is None:
            return
        
        pk_list = self.get_primary_key_names(parent_cls)
        
        query = self.session.query(target_cls) \
            .join(getattr(parent_cls, fkey)) \
            .filter(
                *[
                    getattr(parent_cls, key)==getattr(parent, key)
                    for key in pk_list
                ],
                target_filters)

        related_list = getattr(parent, fkey)
        for child in query:
            related_list.remove(child)

        
    #--- methods for tests -----------------------------------------------------
    
    def get_all_models(self, model_cls, sort_key=None):
        models = list(self.session.query(model_cls).all())
        if sort_key:
            models.sort(key=sort_key)
        return models


from save_to_db.exceptions import MultipleItemsMatch


def match_items_to_models(db_adapter, items_and_fkeys, models):
    """ Matches items to models and returns two lists, one with items and
    another with models. Items in first list match with models from the
    second with the same index in the list.
     
    :param db_adapter: database adapter.
    :param items_and_fkeys: List of items and thei foreign key models.
    :param models: List of ORM models.
    :returns: Lists of matched items and list of matched models for each
        item (returns a list and a list of lists).
    """
    if not items_and_fkeys or not models:
        return [], []
    
    matched_items = []
    matched_models = []  # list of lists
     
    def add_matched(item, model):
        if not item.allow_multi_update and  item in matched_items:
            index = matched_items.index(item)
            other_models = matched_models[index]
            if model == other_models[0]:
                return  # Some ORMs can return same model twice
            matched_models[index] = [
                item.resolve_model([model, other_models[0]])
            ]
            return
        
        for i, model_list in enumerate(matched_models):
            if model in model_list:
                other_item = matched_items[i]
                if item is not other_item:
                    raise MultipleItemsMatch(model, item, other_item)
         
        if item not in matched_items:
            matched_items.append(item)
            matched_models.append([model])
        else:
            index = matched_items.index(item)
            matched_models[index].append(model)
     
    for item, fkeys in items_and_fkeys:
        getters = item.getters
        relations = item.relations
        
        for model in models:
            for getter_fields in getters:
                same = True
                for field_name in getter_fields:
                    if field_name not in item:
                        same = False
                        break
                     
                    if field_name in item.fields:
                        model_field_value = getattr(model, field_name)
                        if item[field_name] != model_field_value:
                            same = False
                            break
                    elif field_name in relations:
                        if field_name not in fkeys:  # not persisted
                            same = False
                            break
                             
                        relation = relations[field_name]
                        if not relation['relation_type'].is_x_to_many():
                            if hasattr(model, field_name):
                                model_field_value = getattr(model, field_name)
                            else:
                                model_field_value = None
                            if fkeys[field_name][0] != model_field_value:
                                same = False
                                break
                        else:
                            contained_models = \
                                db_adapter.related_x_to_many_contains(
                                    model, field_name, fkeys[field_name])
                            if not contained_models:
                                same = False
                                break
                
                if same:
                    break
                     
            if not same:
                continue
            
            add_matched(item, model)
     
    return matched_items, matched_models


from save_to_db.core.item import Item
from save_to_db.utils.test_base import TestBase



class TestNorewrite(TestBase):
    
    ModelGeneralOne = None
    ModelGeneralTwo = None
    
    
    def test_norewrite_fields(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = [{'f_integer'}]
            getters = [{'f_integer'}]
            norewrite_fields = {
                'f_text': True,
                'f_string': False,
            }
         
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            creators = [{'f_integer'}]
            getters = [{'f_integer'}]
            

        persister = self.persister
         
        #--- normal fields ---
        # all fields set
        item_one = ItemGeneralOne(f_integer=1,
                                  f_text='text-1',
                                  f_string='string-1',
                                  f_float=1.0)
        _, model_ones_list = persister.persist(item_one)
        self.assertEqual(len(model_ones_list), 1)
        self.assertEqual(len(model_ones_list[0]), 1)
        model_one = model_ones_list[0][0]
          
        # model did not exist, everything was set
        self.assertEqual(model_one.f_integer, 1)
        self.assertEqual(model_one.f_text, 'text-1')
        self.assertEqual(model_one.f_string, 'string-1')
        self.assertEqual(model_one.f_float, 1.0)
          
        item_one = ItemGeneralOne(f_integer=1,
                                  f_text='text-2',
                                  f_string='string-2',
                                  f_float=2.0)
        _, model_ones_list = persister.persist(item_one)
        self.assertEqual(len(model_ones_list), 1)
        self.assertEqual(len(model_ones_list[0]), 1)
        model_one = model_ones_list[0][0]
          
        # only field that can be rewritten changed
        self.assertEqual(model_one.f_integer, 1)
        self.assertEqual(model_one.f_text, 'text-1')
        self.assertEqual(model_one.f_string, 'string-1')
        self.assertEqual(model_one.f_float, 2.0)  # changed
          
        # model did not exist, fields from norewrite are not set
        item_one = ItemGeneralOne(f_integer=2,
                                  f_float=2.0)
        _, model_ones_list = persister.persist(item_one)
        self.assertEqual(len(model_ones_list), 1)
        self.assertEqual(len(model_ones_list[0]), 1)
        model_one = model_ones_list[0][0]
         
        # model did not exist, everything that item had was set
        self.assertEqual(model_one.f_integer, 2)
        self.assertEqual(model_one.f_text, None)
        self.assertEqual(model_one.f_string, None)
        self.assertEqual(model_one.f_float, 2.0)
         
        item_one = ItemGeneralOne(f_integer=2,
                                  f_text='text-2',
                                  f_string='string-2',
                                  f_float=3.0)
        _, model_ones_list = persister.persist(item_one)
        self.assertEqual(len(model_ones_list), 1)
        self.assertEqual(len(model_ones_list[0]), 1)
        model_one = model_ones_list[0][0]
         
        # only field that can be rewritten changed
        self.assertEqual(model_one.f_integer, 2)
        self.assertEqual(model_one.f_text, 'text-2')
        self.assertEqual(model_one.f_string, None)  # cannot overwrite
        self.assertEqual(model_one.f_float, 3.0)  # changed
    
    
    def test_norewrite_cannot_create_1_to_1(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = [{'f_integer'}]
            getters = [{'f_integer'}]
            norewrite_fields = {
                'two_1_1': False,
            }
         
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            creators = [{'f_integer', 'one_1_1'}]
            getters = [{'f_integer'}]
            
        persister = self.persister
        
        item_one = ItemGeneralOne(f_integer=1)
        persister.persist(item_one)
        
        # item ones' two_1_1 cannot be ovewritten, hence `item_two` cannot
        # be created
        item_two = ItemGeneralTwo(
            f_integer=2,
            one_1_1=item_one
        )
        _, model_lists = persister.persist(item_two)
        self.assertEqual(len(model_lists), 0)
        
        model_twos = self.get_all_models(self.ModelGeneralTwo)
        self.assertEqual(len(model_twos), 0)
        
        model_ones = self.get_all_models(self.ModelGeneralOne)
        self.assertEqual(len(model_ones), 1)
        model_one = model_ones[0]
        self.assertEqual(model_one.f_integer, 1)
        self.assertTrue(not hasattr(model_one, 'one_1_1') or
                        model_one.one_1_1 is None)
    
    
    def test_norewrite_nullables(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = [{'f_integer'}]
            getters = [{'f_integer'}]
            norewrite_fields = {
                'two_1_1': True,
                'two_x_x': True,
            }
            nullables = ['f_text', 'two_1_1', 'two_x_x']
            relations = {
                'two_x_x': {
                    'replace_x_to_many': True,
                }
            }
         
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            creators = [{'f_integer'}]
            getters = [{'f_integer'}]
            nullables = ['f_text', 'one_1_1', 'one_x_x']
            relations = {
                'one_x_x': {
                    'replace_x_to_many': True,
                }
            }
        
        persister = self.persister
        
        item_one = ItemGeneralOne(
            f_integer=1,
            f_text='text',
            two_1_1=ItemGeneralTwo(f_integer=2),
            two_x_x=[ItemGeneralTwo(f_integer=3)]
        )
        _, model_lists = persister.persist(item_one)
        self.assertEqual(len(model_lists), 1)
        self.assertEqual(len(model_lists[0]), 1)
        model_one = model_lists[0][0]
        self.assertEqual(model_one.f_integer, 1)
        self.assertEqual(model_one.f_text, 'text')
        self.assertEqual(model_one.two_1_1.f_integer, 2)
        two_x_x = self.get_related_x_to_many(model_one, 'two_x_x')
        self.assertEqual(len(two_x_x), 1)
        self.assertEqual(two_x_x[0].f_integer, 3)
        
        # not all fields must not be rewritten with `None`
        item_one = ItemGeneralOne(f_integer=1)
        persister.persist(item_one)
        model_ones = self.get_all_models(self.ModelGeneralOne)
        self.assertEqual(len(model_ones), 1)
        
        model_one = model_ones[0]
        self.assertEqual(model_one.f_integer, 1)
        self.assertEqual(model_one.f_text, None)  # rewritten
        self.assertEqual(model_one.two_1_1.f_integer, 2)
        two_x_x = self.get_related_x_to_many(model_one, 'two_x_x')
        self.assertEqual(len(two_x_x), 1)
        self.assertEqual(two_x_x[0].f_integer, 3)
    
    
    def test_norewrite_relations_false(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = [{'f_integer'}]
            getters = [{'f_integer'}]
            norewrite_fields = {
                'two_1_1': False,
                'two_1_x': False,
                'two_x_1': False,
                'two_x_x': False,
                
            }
         
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            creators = [{'f_integer'}]
            getters = [{'f_integer'}]
        
        persister = self.persister
        sort_func = lambda model: model.f_integer
        
        item_one = ItemGeneralOne(f_integer=1)
        persister.persist(item_one)
        model_ones = self.get_all_models(self.ModelGeneralOne, sort_func)
        self.assertEqual(len(model_ones), 1)
        self.assertEqual(model_ones[0].f_integer, 1)
        
        #--- all newly created ---
        item_one = ItemGeneralOne(
            f_integer=1,
            two_1_1=ItemGeneralTwo(f_integer=2),  # cannot rewrite
            two_1_x=[ItemGeneralTwo(f_integer=3),
                     ItemGeneralTwo(f_integer=4)],  # can rewrite
            two_x_1=ItemGeneralTwo(f_integer=5),  # cannot rewrite
            two_x_x=[ItemGeneralTwo(f_integer=6),
                     ItemGeneralTwo(f_integer=7)],  # can rewrite
        )
        _, models_list = persister.persist(item_one)
        self.assertEqual(len(models_list), 1)
        self.assertEqual(len(models_list[0]), 1)
        model_one = models_list[0][0]
        
        # no related models
        two_1_x = self.get_related_x_to_many(model_one, 'two_1_x', sort_func)
        two_x_x = self.get_related_x_to_many(model_one, 'two_x_x', sort_func)
        self.assertEqual(len(two_1_x), 2)
        self.assertEqual(len(two_x_x), 2)
        
        self.assertEqual(two_1_x[0].f_integer, 3)
        self.assertEqual(two_1_x[1].f_integer, 4)
        self.assertEqual(two_x_x[0].f_integer, 6)
        self.assertEqual(two_x_x[1].f_integer, 7)
        
        self.assertEqual(model_one.f_integer, 1)
        self.assertTrue(not hasattr(model_one, 'two_1_1') or
                        model_one.two_1_1 is None)
        self.assertTrue(not hasattr(model_one, 'two_x_1') or
                        model_one.two_x_1 is None)
        
        # but the other models were created
        models_two = self.get_all_models(self.ModelGeneralTwo, sort_func)
        expected_f_integers = [2, 3, 4, 5, 6, 7]
        f_integers = [model.f_integer for model in models_two]
        self.assertEqual(f_integers, expected_f_integers)
        
        #--- all newly created, even model_one ---
        item_one = ItemGeneralOne(
            f_integer=2,
            two_1_1=ItemGeneralTwo(f_integer=1002),
            two_1_x=[ItemGeneralTwo(f_integer=1003),
                     ItemGeneralTwo(f_integer=1004)],
            two_x_1=ItemGeneralTwo(f_integer=1005),
            two_x_x=[ItemGeneralTwo(f_integer=1006),
                     ItemGeneralTwo(f_integer=1007)],
        )
        _, models_list = persister.persist(item_one)
        self.assertEqual(len(models_list), 1)
        self.assertEqual(len(models_list[0]), 1)
        model_one = models_list[0][0]
        
        # relations were set
        two_1_x = self.get_related_x_to_many(model_one, 'two_1_x', sort_func)
        two_x_x = self.get_related_x_to_many(model_one, 'two_x_x', sort_func)
        self.assertEqual(len(two_1_x), 2)
        self.assertEqual(len(two_x_x), 2)
        
        self.assertEqual(model_one.f_integer, 2)
        self.assertEqual(model_one.two_1_1.f_integer, 1002)
        self.assertEqual(two_1_x[0].f_integer, 1003)
        self.assertEqual(two_1_x[1].f_integer, 1004)
        self.assertEqual(model_one.two_x_1.f_integer, 1005)
        self.assertEqual(two_x_x[0].f_integer, 1006)
        self.assertEqual(two_x_x[1].f_integer, 1007)
        
        #--- checking for the other side ---
        item_two = ItemGeneralTwo(f_integer=1000000)
        persister.persist(item_two)
        model_twos = self.get_all_models(self.ModelGeneralTwo, sort_func)
        self.assertEqual(model_twos[-1].f_integer, 1000000)
        
        model_ones = self.get_all_models(self.ModelGeneralOne, sort_func)
        # model one with `f_integer=1000000` does not exists
        self.assertLess(model_ones[-1].f_integer, 1000000)
        item_one = ItemGeneralOne(
            f_integer=1000000,
            two_1_1=item_two,  # cannot rewrite model_two
            two_1_x=[item_two],  # cannot rewrite model_two
            two_x_1=item_two,  # can set
            two_x_x=[item_two]  # can set
        )
        _, models_list = persister.persist(item_one)
        self.assertEqual(len(models_list), 1)
        self.assertEqual(len(models_list[0]), 1)
        model_one = models_list[0][0]
        
        two_1_x = self.get_related_x_to_many(model_one, 'two_1_x', sort_func)
        two_x_x = self.get_related_x_to_many(model_one, 'two_x_x', sort_func)
        self.assertEqual(len(two_1_x), 0)
        self.assertEqual(len(two_x_x), 1)
        self.assertEqual(two_x_x[0].f_integer, 1000000)
        
        self.assertEqual(model_one.f_integer, 1000000)
        self.assertTrue(not hasattr(model_one, 'two_1_1') or
                        model_one.two_1_1 is None)
        self.assertEqual(model_one.two_x_1.f_integer, 1000000)
    
    
    def test_norewrite_relations_true(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = [{'f_integer'}]
            getters = [{'f_integer'}]
            norewrite_fields = {
                'two_1_1': True,
                'two_1_x': True,
                'two_x_1': True,
                'two_x_x': True,
                
            }
         
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            creators = [{'f_integer'}]
            getters = [{'f_integer'}]
        
        persister = self.persister
        sort_func = lambda model: model.f_integer
        
        item_one = ItemGeneralOne(f_integer=1)
        persister.persist(item_one)
        model_ones = self.get_all_models(self.ModelGeneralOne, sort_func)
        self.assertEqual(len(model_ones), 1)
        self.assertEqual(model_ones[0].f_integer, 1)
        
        #--- all newly created that can rewrite none ---
        item_one = ItemGeneralOne(
            f_integer=1,
            two_1_1=ItemGeneralTwo(f_integer=2),
            two_1_x=[ItemGeneralTwo(f_integer=3),
                     ItemGeneralTwo(f_integer=4)],
            two_x_1=ItemGeneralTwo(f_integer=5),
            two_x_x=[ItemGeneralTwo(f_integer=6),
                     ItemGeneralTwo(f_integer=7)],
        )
        _, models_list = persister.persist(item_one)
        self.assertEqual(len(models_list), 1)
        self.assertEqual(len(models_list[0]), 1)
        model_one = models_list[0][0]
        
        two_1_x = self.get_related_x_to_many(model_one, 'two_1_x', sort_func)
        two_x_x = self.get_related_x_to_many(model_one, 'two_x_x', sort_func)
        self.assertEqual(len(two_1_x), 2)
        self.assertEqual(len(two_x_x), 2)
        
        self.assertEqual(model_one.f_integer, 1)
        self.assertEqual(model_one.two_1_1.f_integer, 2)
        self.assertEqual(two_1_x[0].f_integer, 3)
        self.assertEqual(two_1_x[1].f_integer, 4)
        self.assertEqual(model_one.two_x_1.f_integer, 5)
        self.assertEqual(two_x_x[0].f_integer, 6)
        self.assertEqual(two_x_x[1].f_integer, 7)
        
        #--- item_one already exists and has values set, ---
        #--- nothing must be rewritten when all models already exist ---
        bulk_two = ItemGeneralTwo.Bulk()
        for f_integer in range(102, 108):
            bulk_two.gen(f_integer=f_integer)
        persister.persist(bulk_two)
        
        item_one = ItemGeneralOne(
            f_integer=1,
            two_1_1=ItemGeneralTwo(f_integer=102),
            two_1_x=[ItemGeneralTwo(f_integer=103),
                     ItemGeneralTwo(f_integer=104)],
            two_x_1=ItemGeneralTwo(f_integer=105),
            two_x_x=[ItemGeneralTwo(f_integer=106),
                     ItemGeneralTwo(f_integer=107)]
        )

        _, models_list = persister.persist(item_one)
        self.assertEqual(len(models_list), 1)
        self.assertEqual(len(models_list[0]), 1)
        model_one = models_list[0][0]
        
        # nothing changed
        two_1_x = self.get_related_x_to_many(model_one, 'two_1_x', sort_func)
        two_x_x = self.get_related_x_to_many(model_one, 'two_x_x', sort_func)
        self.assertEqual(len(two_1_x), 2)
        self.assertEqual(len(two_x_x), 2)
        
        self.assertEqual(model_one.f_integer, 1)
        self.assertEqual(model_one.two_1_1.f_integer, 2)
        self.assertEqual(two_1_x[0].f_integer, 3)
        self.assertEqual(two_1_x[1].f_integer, 4)
        self.assertEqual(model_one.two_x_1.f_integer, 5)
        self.assertEqual(two_x_x[0].f_integer, 6)
        self.assertEqual(two_x_x[1].f_integer, 7)
        
        #--- existed side "many" can be added to with created models ---
        item_one = ItemGeneralOne(
            f_integer=1,
            two_1_1=ItemGeneralTwo(f_integer=202),
            two_1_x=[ItemGeneralTwo(f_integer=203),
                     ItemGeneralTwo(f_integer=204)],
            two_x_1=ItemGeneralTwo(f_integer=205),
            two_x_x=[ItemGeneralTwo(f_integer=206),
                     ItemGeneralTwo(f_integer=207)]
        )
        _, models_list = persister.persist(item_one)
        self.assertEqual(len(models_list), 1)
        self.assertEqual(len(models_list[0]), 1)
        model_one = models_list[0][0]
        
        # x-to-many were added to and x-to-1 unchanged
        two_1_x = self.get_related_x_to_many(model_one, 'two_1_x', sort_func)
        two_x_x = self.get_related_x_to_many(model_one, 'two_x_x', sort_func)
        self.assertEqual(len(two_1_x), 4)
        self.assertEqual(len(two_x_x), 4)
        
        self.assertEqual(model_one.f_integer, 1)
        self.assertEqual(model_one.two_1_1.f_integer, 2)
        self.assertEqual(two_1_x[0].f_integer, 3)
        self.assertEqual(two_1_x[1].f_integer, 4)
        self.assertEqual(two_1_x[2].f_integer, 203)
        self.assertEqual(two_1_x[3].f_integer, 204)
        self.assertEqual(model_one.two_x_1.f_integer, 5)
        self.assertEqual(two_x_x[0].f_integer, 6)
        self.assertEqual(two_x_x[1].f_integer, 7)
        self.assertEqual(two_x_x[2].f_integer, 206)
        self.assertEqual(two_x_x[3].f_integer, 207)
        
        #--- checking for the other side ---
        item_two = ItemGeneralTwo(f_integer=1000)
        persister.persist(item_two)
        model_twos = self.get_all_models(self.ModelGeneralTwo, sort_func)
        self.assertEqual(model_twos[-1].f_integer, 1000)
        
        model_ones = self.get_all_models(self.ModelGeneralOne, sort_func)
        # model one with `f_integer=1000` does not exists
        self.assertLess(model_ones[-1].f_integer, 1000)
        # model_two can be rewritten
        item_one = ItemGeneralOne(
            f_integer=1000,
            two_1_1=item_two,
            two_1_x=[item_two],
            two_x_1=item_two,
            two_x_x=[item_two]
        )
        _, models_list = persister.persist(item_one)
        self.assertEqual(len(models_list), 1)
        self.assertEqual(len(models_list[0]), 1)
        model_one = models_list[0][0]
        
        two_1_x = self.get_related_x_to_many(model_one, 'two_1_x', sort_func)
        two_x_x = self.get_related_x_to_many(model_one, 'two_x_x', sort_func)
        self.assertEqual(len(two_1_x), 1)
        self.assertEqual(len(two_x_x), 1)
        
        self.assertEqual(model_one.f_integer, 1000)
        self.assertEqual(model_one.two_1_1.f_integer, 1000)
        self.assertEqual(two_1_x[0].f_integer, 1000)
        self.assertEqual(model_one.two_x_1.f_integer, 1000)
        self.assertEqual(two_x_x[0].f_integer, 1000)
        
        #--- no rewrite for model_two ---
        model_ones = self.get_all_models(self.ModelGeneralOne, sort_func)
        # model one with `f_integer=2000` does not exists
        self.assertLess(model_ones[-1].f_integer, 2000)
        item_one = ItemGeneralOne(f_integer=2000)
        persister.persist(item_one)
        
        # model_two cannot be rewritten
        item_one = ItemGeneralOne(
            f_integer=2000,
            two_1_1=item_two,
            two_1_x=[item_two],
            two_x_1=item_two,
            two_x_x=[item_two]
        )
        _, models_list = persister.persist(item_one)
        self.assertEqual(len(models_list), 1)
        self.assertEqual(len(models_list[0]), 1)
        model_one = models_list[0][0]
        
        two_1_x = self.get_related_x_to_many(model_one, 'two_1_x', sort_func)
        two_x_x = self.get_related_x_to_many(model_one, 'two_x_x', sort_func)
        self.assertEqual(len(two_1_x), 0)
        self.assertEqual(len(two_x_x), 0)
        
        self.assertEqual(model_one.f_integer, 2000)
        self.assertTrue(not hasattr(model_one, 'two_1_1') or
                        model_one.two_1_1 is None)
        self.assertTrue(not hasattr(model_one, 'two_x_1') or
                        model_one.two_x_1 is None)
    
        # model one with `f_integer=3000` does not exists
        self.assertLess(model_ones[-1].f_integer, 3000)
        
        # "many" side can be added to using new models if the side existed
        item_one = ItemGeneralOne(
            f_integer=3000,
            two_1_1=item_two,  # cannot rewrite item two
            two_1_x=[item_two],  # cannot redirect item_two
            two_x_1=item_two,  # can add to "many" side
            two_x_x=[item_two]  # can add to "many" side
        )
        _, models_list = persister.persist(item_one)
        self.assertEqual(len(models_list), 1)
        self.assertEqual(len(models_list[0]), 1)
        model_one = models_list[0][0]
        
        two_1_x = self.get_related_x_to_many(model_one, 'two_1_x', sort_func)
        two_x_x = self.get_related_x_to_many(model_one, 'two_x_x', sort_func)
        self.assertEqual(len(two_1_x), 0)
        self.assertEqual(len(two_x_x), 1)
        self.assertEqual(two_x_x[0].f_integer, 1000)
        
        self.assertEqual(model_one.f_integer, 3000)
        self.assertTrue(not hasattr(model_one, 'two_1_1') or
                        model_one.two_1_1 is None)
        self.assertEqual(model_one.two_x_1.f_integer, 1000)
        
    
    def test_norewrite_none(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = [{'f_integer'}]
            getters = [{'f_integer'}]
            norewrite_fields = {
                'f_string': None,
                'two_1_1': None,
                True: True,
            }
          
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            creators = [{'f_integer'}]
            getters = [{'f_integer'}]
         
        persister = self.persister
         
        # creating
        item_one = ItemGeneralOne(f_integer=1,
                                  # cannot rewrite
                                  f_string='str-1',
                                  two_1_1=ItemGeneralTwo(f_integer=102),
                                  # can rewrite
                                  f_text='text-1',
                                  two_x_1=ItemGeneralTwo(f_integer=103))
        persister.persist(item_one)
        model_ones = self.get_all_models(self.ModelGeneralOne)
        self.assertEqual(len(model_ones), 1)
        model_one = model_ones[0]
        self.assertEqual(model_one.f_integer, 1)
        self.assertEqual(model_one.f_string, 'str-1')
        self.assertIsNotNone(model_one.two_1_1)
        self.assertEqual(model_one.two_1_1.f_integer, 102)
        self.assertEqual(model_one.f_text, 'text-1')
        self.assertEqual(model_one.two_x_1.f_integer, 103)
         
        # updating
        item_one = ItemGeneralOne(f_integer=1,
                                  # cannot rewrite
                                  f_string='str-2',
                                  two_1_1=ItemGeneralTwo(f_integer=202),
                                  # can rewrite
                                  f_text='text-2',
                                  two_x_1=ItemGeneralTwo(f_integer=203))

        persister.persist(item_one)
        model_ones = self.get_all_models(self.ModelGeneralOne)
        self.assertEqual(len(model_ones), 1)
        model_one = model_ones[0]
        self.assertEqual(model_one.f_integer, 1)
        self.assertEqual(model_one.f_string, 'str-2')
        self.assertIsNotNone(model_one.two_1_1)
        self.assertEqual(model_one.two_1_1.f_integer, 202)
        self.assertEqual(model_one.f_text, 'text-1')
        self.assertEqual(model_one.two_x_1.f_integer, 103)
        
    
    def test_norewrite_false_creation(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = [{'f_integer'}]
            getters = [{'f_integer'}]
            norewrite_fields = {
                'two_1_1': False,
                'two_1_x': False,
                'two_x_1': False,
                'two_x_x': False,
            }
          
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            creators = [{'f_integer'}]
            getters = [{'f_integer'}]
        
        sort_func = lambda model: model.f_integer
        
        model_ones = self.get_all_models(self.ModelGeneralOne)
        model_twos = self.get_all_models(self.ModelGeneralTwo)
        self.assertEqual(len(model_ones), 0)
        self.assertEqual(len(model_twos), 0)    
        
        item_one = ItemGeneralOne(f_integer=1,
                                  two_1_1=ItemGeneralTwo(f_integer=1),
                                  two_1_x=[ItemGeneralTwo(f_integer=2)],
                                  two_x_1=ItemGeneralTwo(f_integer=3),
                                  two_x_x=[ItemGeneralTwo(f_integer=4)],)
        self.persister.persist(item_one)
        
        model_ones = self.get_all_models(self.ModelGeneralOne,
                                         sort_key=sort_func)
        model_twos = self.get_all_models(self.ModelGeneralTwo,
                                         sort_key=sort_func)
        # everything can be created
        self.assertEqual(len(model_ones), 1)
        self.assertEqual(len(model_twos), 4)
        
        two_1_x = self.get_related_x_to_many(model_ones[0], 'two_1_x',
                                             sort_key=sort_func)
        two_x_x = self.get_related_x_to_many(model_ones[0], 'two_x_x',
                                             sort_key=sort_func)
        self.assertEqual(len(two_1_x), 1)
        self.assertEqual(len(two_x_x), 1)
        self.assertEqual(model_ones[0].f_integer, 1)
        self.assertEqual(two_1_x[0].f_integer, 2)
        self.assertEqual(model_ones[0].two_x_1.f_integer, 3)
        self.assertEqual(two_x_x[0].f_integer, 4)
        
        # case when all models already existed
        bulk_two = ItemGeneralTwo.Bulk()
        bulk_two.gen(f_integer=5)
        bulk_two.gen(f_integer=6)
        bulk_two.gen(f_integer=7)
        bulk_two.gen(f_integer=8)
        self.persister.persist(bulk_two)
        # updating
        item_one = ItemGeneralOne(f_integer=1,
                                  two_1_1=ItemGeneralTwo(f_integer=5),
                                  two_1_x=[ItemGeneralTwo(f_integer=6)],
                                  two_x_1=ItemGeneralTwo(f_integer=7),
                                  two_x_x=[ItemGeneralTwo(f_integer=8)],)
        self.persister.persist(item_one)
        
        model_ones = self.get_all_models(self.ModelGeneralOne,
                                         sort_key=sort_func)
        model_twos = self.get_all_models(self.ModelGeneralTwo,
                                         sort_key=sort_func)
        self.assertEqual(len(model_ones), 1)
        self.assertEqual(len(model_twos), 8)  # models two existed
        
        # but model one stayed the same
        two_1_x = self.get_related_x_to_many(model_ones[0], 'two_1_x',
                                             sort_key=sort_func)
        two_x_x = self.get_related_x_to_many(model_ones[0], 'two_x_x',
                                             sort_key=sort_func)
        self.assertEqual(len(two_1_x), 1)
        self.assertEqual(len(two_x_x), 1)
        self.assertEqual(model_ones[0].f_integer, 1)
        self.assertEqual(two_1_x[0].f_integer, 2)
        self.assertEqual(model_ones[0].two_x_1.f_integer, 3)
        self.assertEqual(two_x_x[0].f_integer, 4)
        
        # case when models two are created
        item_one = ItemGeneralOne(f_integer=1,
                                  two_1_1=ItemGeneralTwo(f_integer=9),
                                  two_1_x=[ItemGeneralTwo(f_integer=10)],
                                  two_x_1=ItemGeneralTwo(f_integer=11),
                                  two_x_x=[ItemGeneralTwo(f_integer=12)],)
        self.persister.persist(item_one)
        
        model_ones = self.get_all_models(self.ModelGeneralOne,
                                         sort_key=sort_func)
        model_twos = self.get_all_models(self.ModelGeneralTwo,
                                         sort_key=sort_func)
        # everything can be created
        self.assertEqual(len(model_ones), 1)
        self.assertEqual(len(model_twos), 12)  # models two created
        
        # but model one stayed the same
        two_1_x = self.get_related_x_to_many(model_ones[0], 'two_1_x',
                                             sort_key=sort_func)
        two_x_x = self.get_related_x_to_many(model_ones[0], 'two_x_x',
                                             sort_key=sort_func)
        # x-to-many updated, but x-to-1 not
        self.assertEqual(len(two_1_x), 2)
        self.assertEqual(len(two_x_x), 2)
        self.assertEqual(model_ones[0].f_integer, 1)
        self.assertEqual(two_1_x[0].f_integer, 2)
        self.assertEqual(two_1_x[1].f_integer, 10)
        self.assertEqual(model_ones[0].two_x_1.f_integer, 3)
        self.assertEqual(two_x_x[0].f_integer, 4)
        self.assertEqual(two_x_x[1].f_integer, 12)
        
        
    def test_norewrite_false_cannot_create_1_x(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = [{'f_integer'}]
            getters = [{'f_integer'}]
            norewrite_fields = {
                'two_1_1': False,
                'two_1_x': False,
                'two_x_1': False,
                'two_x_x': False,
            }
         
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            creators = [{'one_1_1', 'one_1_x', 'f_integer',}]
            getters = [{'f_integer',}]
        
        sort_func = lambda model: model.f_integer
        
        model_ones = self.get_all_models(self.ModelGeneralOne)
        model_twos = self.get_all_models(self.ModelGeneralTwo)
        self.assertEqual(len(model_ones), 0)
        self.assertEqual(len(model_twos), 0)    
        
        # everything can be created
        item_two = ItemGeneralTwo(f_integer=2)
        item_one = ItemGeneralOne(f_integer=1,
                                  two_1_1=item_two,
                                  two_x_1=item_two)
        self.persister.persist(item_one)
        
        model_ones = self.get_all_models(self.ModelGeneralOne,
                                         sort_key=sort_func)
        model_twos = self.get_all_models(self.ModelGeneralTwo,
                                         sort_key=sort_func)
        
        self.assertEqual(len(model_ones), 1)
        self.assertEqual(len(model_twos), 1)
        
        self.assertEqual(model_ones[0].f_integer, 1)
        self.assertEqual(model_ones[0].two_1_1.f_integer, 2)
        self.assertEqual(model_ones[0].two_x_1.f_integer, 2)
        
        # models twos cannot be created (cannot rewrite model_one)
        item_two = ItemGeneralTwo(f_integer=3)
        item_one = ItemGeneralOne(f_integer=1,
                                  two_1_1=item_two,
                                  two_x_1=item_two)
        self.persister.persist(item_one)
        
        # nothing changed
        model_ones = self.get_all_models(self.ModelGeneralOne,
                                         sort_key=sort_func)
        model_twos = self.get_all_models(self.ModelGeneralTwo,
                                         sort_key=sort_func)
        
        self.assertEqual(len(model_ones), 1)
        self.assertEqual(len(model_twos), 1)
        
        self.assertEqual(model_ones[0].f_integer, 1)
        self.assertEqual(model_ones[0].two_1_1.f_integer, 2)
        self.assertEqual(model_ones[0].two_x_1.f_integer, 2)


from django.db import models


class CommonColumns(models.Model):
    class Meta:
        abstract = True
        
    f_binary = models.BinaryField(null=True)
    f_boolean = models.NullBooleanField(null=True)
    f_string = models.CharField(max_length=255, null=True)
    f_text = models.TextField(null=True)
    f_integer = models.IntegerField(null=True)
    f_float = models.FloatField(null=True)
    f_date = models.DateField(null=True)
    f_time = models.TimeField(null=True)
    f_datetime = models.DateTimeField(null=True)
    
    
    
class ModelConstraintsOne(CommonColumns):
    class Meta:
        db_table = 'model_constraints_one'
        
    f_text = models.TextField(null=False)
    f_integer = models.IntegerField(null=True, unique=True)
    f_string = models.CharField(max_length=255, null=False, unique=True)
    
    # `two_1_x`, `three_1_x` and `five_1_1` configured on the other side
        


class ModelConstraintsTwo(CommonColumns):
    class Meta:
        db_table = 'model_constraints_two'
        
    one_x_1 = models.ForeignKey(
        'ModelConstraintsOne',
        on_delete=models.CASCADE,
        related_name='two_1_x',
        null=True)
    
    # Django does not support composite primary key
    four_x_1 = models.ForeignKey(
        'ModelConstraintsFour',
        on_delete=models.CASCADE,
        related_name='two_1_x',
        null=False)
    
    # `three_x_x` configured on the other side
    
    

class ModelConstraintsThree(CommonColumns):
    class Meta:
        db_table = 'model_constraints_three'
        
    one_x_1 = models.ForeignKey(
        'ModelConstraintsOne',
        on_delete=models.CASCADE,
        related_name='three_1_x',
        null=False)
    
    two_x_x = models.ManyToManyField(
        'ModelConstraintsTwo',
        related_name='three_x_x')
    
    four_x_1 = models.ForeignKey(
        'ModelConstraintsFour',
        on_delete=models.CASCADE,
        related_name='three_1_x',
        null=True)
    


class ModelConstraintsFour(CommonColumns):
    class Meta:
        db_table = 'model_constraints_four'
        unique_together = (('f_integer', 'f_string'),)
        
    # Django does not support composite primary key, it uses `id` by default

    # `two_1_x`, `three_1_x`, `five_1_1` configured on the other side



class ModelConstraintsFive(CommonColumns):
    class Meta:
        db_table = 'model_constraints_five'
    
    one_1_1 = models.OneToOneField(
        'ModelConstraintsOne',
        on_delete=models.CASCADE,
        related_name='five_1_1',
        null=True)
    
    four_1_1 = models.OneToOneField(
        'ModelConstraintsFour',
        on_delete=models.CASCADE,
        related_name='five_1_1',
        null=True)
    
    # `six_1_1` configured on the other side



class ModelConstraintsSix(CommonColumns):
    class Meta:
        db_table = 'model_constraints_six'
        unique_together = (('five_1_1', 'f_integer'),)
        
    five_1_1 = models.OneToOneField(
        'ModelConstraintsFive',
        on_delete=models.CASCADE,
        related_name='six_1_1',
        null=False)
        
        
        
class ModelConstraintsSelf(CommonColumns):
    class Meta:
        db_table = 'model_constraints_self'
    
    # using non-standard primary key
    code = models.CharField(max_length=255, primary_key=True)
        
    # many-to-one, not_null
    parent_x_1 = models.ForeignKey(
        'ModelConstraintsSelf',
        on_delete=models.CASCADE,
        related_name='child_1_x',
        null=False)
    
    # self one-to-one, unique
    first_parent_1_1 = models.OneToOneField(
        'ModelConstraintsSelf',
        on_delete=models.CASCADE,
        related_name='first_child_1_1',
        null=True)
    
    # self one-to-one, unique, not_null
    second_parent_1_1 = models.OneToOneField(
        'ModelConstraintsSelf',
        on_delete=models.CASCADE,
        related_name='second_child_1_1',
        null=False)
    
    # self many-to-many
    parent_x_x = models.ManyToManyField(
        'ModelConstraintsSelf',
        related_name='child_x_x')



    
    


from save_to_db.exceptions import BulkItemOnetoXDefaultError, WrongAlias
from save_to_db.core.item import Item
from save_to_db.utils.test_base import TestBase



class TestBasicUse(TestBase):
    
    ModelFieldTypes = None
    ModelGeneralOne = None
    ModelGeneralTwo = None
    
    
    @classmethod
    def setup_models(cls, aliased=False):
        cls.item_cls_manager.clear()
        
        class ItemFieldTypes(Item):
            model_cls = cls.ModelFieldTypes
            conversions = {
                'decimal_separator': ',',
            }
        cls.ItemFieldTypes = ItemFieldTypes
        
        dct_1 = {'model_cls': cls.ModelGeneralOne}
        dct_2 = {'model_cls': cls.ModelGeneralTwo}
        
        if aliased:
            dct_1.update({
                'aliases': {
                    'alias_1': 'two_1_1__f_integer',
                    'alias_2': 'two_x_x__one_1_1__f_integer',
                }
            })
            dct_2.update({
                'aliases': {
                    'alias_1__post': 'f_string',
                    'alias_2__post': 'one_x_1__alias_1',
                    'alias_3__post': 'one_x_x__two_1_x__alias_1__post',
                }
            })
        
        cls.ItemGeneralOne = type('ItemGeneralOne', (Item,), dct_1)
        cls.ItemGeneralTwo = type('ItemGeneralTwo', (Item,), dct_2)
        
        cls.ItemGeneralOne.complete_setup()
        cls.ItemGeneralTwo.complete_setup()
    
    
    def test_setting_fields(self):
        self.setup_models()

        #--- setting with correct keys ---
        item = self.ItemGeneralOne()
        item['f_integer'] = '10'
        item['two_1_1__f_string'] = 'one'
        item['two_1_1__one_1_x__f_string'] = 'two'
        item['two_1_1__one_x_x__two_x_1__f_string'] = 'three'
        bulk = item['two_1_1__one_x_x']
        bulk['f_integer'] = '20'
        bulk.gen(f_text='four')
        in_bulk_item = bulk.gen(f_text='five')
        # setting field by calling
        in_bulk_item(f_string='ITEM_CALL') 
        bulk(two_x_1__f_text='BULK_CALL')
        
        expected_value = {
            'item': {
                'f_integer': '10',
                'two_1_1': {
                    'item': {
                        'f_string': 'one',
                        'one_1_x': {
                            'bulk': [],
                            'defaults': {
                                'f_string': 'two'
                            }
                        },
                        'one_x_x': {
                            'bulk': [
                                {
                                    'item': {
                                        'f_text': 'four'
                                    }
                                },
                                {
                                    'item': {
                                        'f_string': 'ITEM_CALL',
                                        'f_text': 'five'
                                    }
                                }
                            ],
                            'defaults': {
                                'f_integer': '20',
                                'two_x_1__f_string': 'three',
                                'two_x_1__f_text': 'BULK_CALL'
                            }
                        }
                    }
                }
            }
        }
        self.assertEqual(item.to_dict(), expected_value)
        
        #--- setting with incorrect keys ---
        item = self.ItemGeneralOne()
        with self.assertRaises(WrongAlias):
            item['wrong_key'] = 'value'
        
        bulk = self.ItemGeneralOne.Bulk()
        with self.assertRaises(WrongAlias):
            bulk['wrong_key'] = 'value'
            
        #--- setting one-to-x default value in bulk
        
        # not items temselves, can belong to different items
        bulk['two_1_1__f_integer'] = '10'
        bulk['two_1_x__f_integer'] = '10'
        bulk['two_x_x__one_1_x__f_integer'] = '10'
         
        # direct
        with self.assertRaises(BulkItemOnetoXDefaultError):
            bulk['two_1_1'] = self.ItemGeneralTwo()
         
        with self.assertRaises(BulkItemOnetoXDefaultError):
            bulk['two_1_x'] = self.ItemGeneralTwo()
         
        with self.assertRaises(BulkItemOnetoXDefaultError):
            bulk['two_1_x'].gen()
         
        with self.assertRaises(BulkItemOnetoXDefaultError):
            bulk['two_x_x__one_1_x'].gen()
        
        #--- aliases -----------------------------------------------------------
        self.setup_models(aliased=True)
        
        item = self.ItemGeneralOne()
        item['alias_1'] = '1'
        item['alias_2'] = '2'
        item['two_1_1__alias_1__post'] = 'str-1'
        item['two_1_1__alias_2__post'] = '3'
        item['two_1_1__alias_3__post'] = '4'
        bulk = item['two_1_x']
        bulk['alias_1__post'] = 'str-2'
        
        expected_value = {
            'id': 1,
            'item': {
                'two_x_x': {
                    'bulk': [],
                    'defaults': {
                        'one_1_1__f_integer': '2'  # alias_1
                    }
                },
                'two_1_1': {
                    'item': {
                        'f_integer': '1',  # alias_1
                        'one_x_1': {
                            'item': {
                                'two_1_1': {
                                    'item': {
                                        # two_1_1__alias_2__post
                                        'f_integer': '3'
                                    }
                                }
                            }
                        },
                        'one_x_x': {
                            'bulk': [],
                            'defaults': {
                                'two_1_x__f_string': '4'
                            }
                        },
                        'f_string': 'str-1'  # two_1_1__alias_1__post
                    }
                },
                'two_1_x': {
                    'bulk': [],
                    'defaults': {
                        'one_x_1': {
                            'id': 1
                        },
                        'f_string': 'str-2'
                    }
                }
            }
        }
        
        self.assertEqual(item.to_dict(), expected_value)
    
    
    def test_getting_fields(self):
        self.setup_models()
        
        item = self.ItemGeneralOne()
        related_item = item['two_x_1__one_1_1__two_1_1'](f_integer='20',
                                                         f_string='value')
        self.assertIsInstance(related_item, self.ItemGeneralTwo)
        
        expected_value = {
            'id': 2,
            'item': {
                'two_x_1': {
                    'id': 1,
                    'item': {
                        'f_integer': '20',
                        'f_string': 'value',
                        'one_1_1': {
                            'item': {
                                'two_1_1': {
                                    'id': 1
                                }
                            }
                        },
                        'one_1_x': {
                            'bulk': [
                                {
                                    'id': 2
                                }
                            ],
                            'defaults': {
                                'two_x_1': {
                                    'id': 1
                                }
                            }
                        }
                    }
                }
            }
        }
        
        self.assertEqual(item.to_dict(), expected_value)
        
        #--- aliases -----------------------------------------------------------
        self.setup_models(aliased=True)
        
        item = self.ItemGeneralOne()
        item['alias_1'] = '1'
        item['alias_2'] = '2'
        item['two_1_1__alias_1__post'] = 'str-1'
        item['two_1_1__alias_2__post'] = '3'
        item['two_1_1__alias_3__post'] = '4'
        bulk = item['two_1_x']
        bulk['alias_1__post'] = 'str-2'
        
        self.assertEqual(item['alias_1'], '1')
        self.assertEqual(item['alias_2'], '2')
        self.assertEqual(item['two_1_1__alias_1__post'], 'str-1')
        self.assertEqual(item['two_1_1__alias_2__post'], '3')
        self.assertEqual(item['two_1_1__alias_3__post'], '4')
        self.assertEqual(bulk['alias_1__post'], 'str-2')
    
    
    def test_getting_fields_reverse_autocomplete(self):
        self.setup_models()
        
        # --- single item ---
        
        # one-to-one
        item_one = self.ItemGeneralOne()
        item_two = item_one['two_1_1']
        self.assertTrue(item_two.is_single_item())
        self.assertIs(item_two['one_1_1'], item_one)
        
        # one-to-many
        item_one = self.ItemGeneralOne()
        bulk_two = item_one['two_1_x']
        self.assertTrue(bulk_two.is_bulk_item())
        self.assertIs(bulk_two['one_x_1'], item_one)  # default value
        
        # many-to-one
        item_one = self.ItemGeneralOne()
        item_two = item_one['two_x_1']
        self.assertTrue(item_two.is_single_item())
        self.assertTrue(item_two['one_1_x'].is_bulk_item())
        # must contain the item
        self.assertEqual(len(item_two['one_1_x']), 1)  # only 1 item
        self.assertIs(item_two['one_1_x'][0], item_one)
        # that item must have a default
        self.assertIs(item_two['one_1_x']['two_x_1'], item_two)
        
        # many-to-many
        item_one = self.ItemGeneralOne()
        bulk_two = item_one['two_x_x']
        self.assertTrue(bulk_two.is_bulk_item())
        self.assertTrue(item_one['two_x_x'].is_bulk_item())
        self.assertEqual(len(bulk_two['one_x_x']), 1)  # only 1 item
        # bulk item must have a default
        self.assertEqual(len(bulk_two['one_x_x']), 1)
        self.assertIs(bulk_two['one_x_x'][0], item_one)
        # but no items two were created
        self.assertEqual(len(bulk_two), 0)
   
    
    def test_del_from_item(self):
        self.setup_models()
        
        item = self.ItemGeneralOne()
        item['f_integer'] = '10'
        item['f_string'] = 'str-10'
        item['two_1_1__f_integer'] = '20'
        item['two_1_1__f_string'] = 'str-20'
        item['two_x_x__one_1_1__f_integer'] = '30'
        item['two_x_x__one_1_1__f_string'] = 'str-30'
        bulk = item['two_1_x']
        bulk['f_string'] = 'str-2'
        
        expect = {
            'item': {
                'two_1_1': {
                    'item': {
                        'f_integer': '20',
                        'f_string': 'str-20'
                    }
                },
                'two_x_x': {
                    'bulk': [],
                    'defaults': {
                        'one_1_1__f_integer': '30',
                        'one_1_1__f_string': 'str-30'
                    }
                },
                'f_integer': '10',
                'f_string': 'str-10',
                'two_1_x': {
                    'bulk': [],
                    'defaults': {
                        'one_x_1': {
                            'id': 1
                        },
                        'f_string': 'str-2'
                    }
                }
            },
            'id': 1
        }

        self.assertEqual(item.to_dict(), expect)
        
        del expect['item']['f_string']
        del item['f_string']
        self.assertEqual(item.to_dict(), expect)
        
        del expect['item']['two_1_1']['item']['f_string']
        del item['two_1_1__f_string']
        self.assertEqual(item.to_dict(), expect)
        
        del expect['item']['two_x_x']['defaults']['one_1_1__f_string']
        del item['two_x_x__one_1_1__f_string']
        self.assertEqual(item.to_dict(), expect)
        
        del bulk['f_string']
        del expect['item']['two_1_x']['defaults']['f_string']
        self.assertEqual(item.to_dict(), expect)
        
        #--- aliases -----------------------------------------------------------
        self.setup_models(aliased=True)
        
        item = self.ItemGeneralOne()
        item['alias_1'] = '1'
        item['alias_2'] = '2'
        item['two_1_1__alias_1__post'] = 'str-1'
        item['two_1_1__alias_2__post'] = '3'
        item['two_1_1__alias_3__post'] = '4'
        bulk = item['two_1_x']
        bulk['alias_1__post'] = 'str-2'
        
        expected_value = {
            'item': {
                'two_1_x': {
                    'defaults': {
                        'f_string': 'str-2',
                        'one_x_1': {
                            'id': 1
                        }
                    },
                    'bulk': []
                },
                'two_x_x': {
                    'defaults': {
                        'one_1_1__f_integer': '2'
                    },
                    'bulk': []
                },
                'two_1_1': {
                    'item': {
                        'f_string': 'str-1',
                        'one_x_1': {
                            'item': {
                                'two_1_1': {
                                    'item': {
                                        'f_integer': '3'
                                    }
                                }
                            }
                        },
                        'one_x_x': {
                            'defaults': {
                                'two_1_x__f_string': '4'
                            },
                            'bulk': []
                        },
                        'f_integer': '1'
                    }
                }
            },
            'id': 1
        }

        self.assertEqual(item.to_dict(), expected_value)
        
        del item['alias_1']
        del expected_value['item']['two_1_1']['item']['f_integer']
        self.assertEqual(item.to_dict(), expected_value)
        
        del item['alias_2']
        del expected_value['item']['two_x_x']['defaults']['one_1_1__f_integer']
        self.assertEqual(item.to_dict(), expected_value)
        
        del item['two_1_1__alias_1__post']
        del expected_value['item']['two_1_1']['item']['f_string']
        self.assertEqual(item.to_dict(), expected_value)
        
        del item['two_1_1__alias_2__post']
        del expected_value['item']['two_1_1']['item']['one_x_1']['item']['two_1_1']['item']['f_integer']
        self.assertEqual(item.to_dict(), expected_value)
        
        del item['two_1_1__alias_3__post']
        del expected_value['item']['two_1_1']['item']['one_x_x']['defaults']['two_1_x__f_string']
        self.assertEqual(item.to_dict(), expected_value)
        
        del bulk['alias_1__post']
        del expected_value['item']['two_1_x']['defaults']['f_string']
        self.assertEqual(item.to_dict(), expected_value)
    
    
    def test_contains(self):
        self.setup_models()
        item = self.ItemGeneralOne()
        item['f_integer'] = '1'
        item['two_1_x__f_integer'] = '2'
        bulk = item['two_1_x']
        bulk['one_x_x__f_integer'] = '3'
        
        self.assertIn('f_integer', item)
        self.assertIn('two_1_x__f_integer', item)
        self.assertIn('two_1_x__one_x_x__f_integer', item)  # from bulk
        
        self.assertNotIn('wrong_key', item)
        self.assertNotIn('wrong_key__f_integer', item)
        self.assertNotIn('two_1_x__one_x_x__wrong_key', item)
        
        self.assertIn('f_integer', bulk)  # item
        self.assertIn('one_x_x__f_integer', bulk)
        
        self.assertNotIn('wrong_key', bulk)
        self.assertNotIn('wrong_key__f_integer', bulk)
        self.assertNotIn('two_1_x__one_x_x__wrong_key', bulk)
        
        #--- aliased -----------------------------------------------------------
        self.setup_models(aliased=True)
        
        item = self.ItemGeneralOne()
        item['alias_1'] = '1'
        item['alias_2'] = '2'
        item['two_1_1__alias_1__post'] = 'str-1'
        item['two_1_1__alias_2__post'] = '3'
        item['two_1_1__alias_3__post'] = '4'
        bulk = item['two_1_x']
        bulk['alias_1__post'] = 'str-2'
        
        self.assertIn('alias_1', item)
        self.assertIn('alias_2', item)
        self.assertIn('two_1_1__alias_1__post', item)
        self.assertIn('two_1_1__alias_2__post', item)
        self.assertIn('two_1_1__alias_3__post', item)
        
        self.assertNotIn('wrong_key', item)
        self.assertNotIn('wrong_key__f_integer', item)
        self.assertNotIn('two_1_x__one_x_x__wrong_key', item)
        
        self.assertIn('alias_1__post', bulk)
        
        self.assertNotIn('wrong_key', bulk)
        self.assertNotIn('wrong_key__f_integer', bulk)
        self.assertNotIn('two_1_x__one_x_x__wrong_key', bulk)
    
    
    def test_bulk_iter(self):
        self.setup_models()
        bulk = self.ItemGeneralOne.Bulk()
        items = [bulk.gen(f_integer=1),
                 bulk.gen(f_integer=2),
                 bulk.gen(f_integer=3),]
        items_in_bulk = list(bulk)
        self.assertEqual(items_in_bulk, items)
        
        for item_no, item_in_bulk in enumerate(bulk):
            self.assertIs(items[item_no], item_in_bulk)
            self.assertIs(items[item_no], bulk[item_no])
        
        self.assertEqual(item_no, len(items)-1)
        
        
    def test_bulk_item_slice(self):
        class ItemsGeneralOne(Item):
            model_cls = self.ModelGeneralOne
                
        class ItemsGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
                
        
        bulk = ItemsGeneralOne.Bulk()
        bulk['f_string'] = 'str-value'
        
        items = []
        for i in range(10):
            item = ItemsGeneralOne(f_integer=i)
            items.append(item)
            bulk.add(item)
        
        first_slice = bulk[0:5]
        self.assertEqual(items[:5], first_slice.bulk)
        self.assertEqual(first_slice['f_string'], 'str-value')
        
        first_slice['f_text'] = 'text-value'
        self.assertNotIn('f_text', bulk)
        
        second_slice = bulk[0::2]
        self.assertEqual(items[0::2], second_slice.bulk)
        self.assertEqual(second_slice['f_string'], 'str-value')
        
        bulk['f_float'] = '10.10'
        self.assertNotIn('f_float', second_slice)
    
    
#     def test_x(self):
#         print('')
#         
#         class Instrument(Item):
#             model_cls = self.ModelGeneralOne
#                 
#         class Dividend(Item):
#             model_cls = self.ModelGeneralTwo
#         
#         dividend = Dividend(f_string='Dividend-1')
#         instrument = dividend['one_x_1']
#         instrument['f_string'] = 'Instrument-1'
#         dividend_2 = instrument['two_1_x'].gen(f_string='Dividend-2')
#         
# #         item_two['f_string'] = 'item_two'
#         
# #         item_two['one_1_x'].gen(f_string='1')
# # #         item_two['one_1_x'].gen(f_string='2')
# #         
# # #         item_one.process()
# #     
# #         print(item_one['two_x_1']['one_1_x'][0]['f_string'])
# #         print(item_one['two_x_1']['one_1_x'][1]['f_string'])
#         
# #         instrument.process()
#         dividend_2.pprint()


from save_to_db.exceptions import CannotMergeModels, MergeModelsNotAllowed
from .relation_type import RelationType


def __one_to_many_merge(db_adapter,
                        result_model, other_model,
                        result_children, other_children,
                        fkey, reverse_key,
                        merge_function, ignore_fields):
    if not result_children or not other_children or not reverse_key:
        return other_children  # no merging needed
    
    # getting unique constraints
    f_model_cls = type(result_children[0])
    unique_constraints = db_adapter.iter_unique_field_combinations(
        f_model_cls)

    unique_constraints = [
        set(field_names)
            for field_names in unique_constraints
            if reverse_key in field_names
    ]
    if not unique_constraints:
        return other_children  # no broken constraints
    
    # looking for conflicting constraints
    def grab_model_values(model, field_names):
        return [
            getattr(model, field_name)
                    if hasattr(model, field_name)
                    else None
                for field_name in field_names
                # parent is supposed to be the same
                if field_name != reverse_key and field_name not in ignore_fields
        ]
    
    for result_i, result_child in enumerate(result_children):
        
        for other_child in other_children[:]:
            broken_constraints = []
            for constraint in unique_constraints:
                outer_values = grab_model_values(result_child, constraint)
                inner_values = grab_model_values(other_child, constraint)
                if outer_values and outer_values == inner_values:
                    broken_constraints.append(constraint)
                    other_children.remove(other_child)
        
            if broken_constraints:
                if merge_function is False:
                    raise MergeModelsNotAllowed(result_model, other_model,
                                                fkey)
                
                new_model = merge_function(result_model, other_model,
                                           result_child, other_child,
                                           fkey, reverse_key,
                                           broken_constraints)
                
                result_children[result_i] = new_model
            
    return other_children


def merge_models(db_adapter, models, merge_policy=None, ignore_fields=None):
    """ See
    :py:meth:`~save_to_db.adapters.utils.adapter_base.AdapterBase.merge_models`
    method of
    :py:class:`~save_to_db.adapters.utils.adapter_base.AdapterBase` class.
    """
    if not models:
        return None
    if len(models) == 1:
        return models[0]
    
    ignore_fields = ignore_fields or []
    all_models = models
    
    # first model is the resulting model
    result_model, *models = all_models
    
    merge_policy = merge_policy or {}
    merge_cls_policy = merge_policy.get(type(result_model), {})
    
    model_cls = type(result_model)
    relations = [
        [fkey, direction, reverse_fkey]
        for fkey, _, direction, reverse_fkey
        in db_adapter.iter_relations(model_cls)
    ]
    fields = [key for key, _ in db_adapter.iter_fields(model_cls)]
    
    # checking for x-to-one relations
    for fkey, direction, reverse_key in relations:
        if fkey in ignore_fields or not direction.is_x_to_one():
            continue

        model_and_fmodels = [
            [model, getattr(model, fkey, None)] for model in all_models
        ]
        
        # making sure models are all the same
        this_model, this_child_model = model_and_fmodels[0]
        result_child_model = this_child_model
        for merged_model, child_merged_model in model_and_fmodels[1:]:
            if child_merged_model is None:
                continue
            if result_child_model is None:
                result_child_model = child_merged_model
                # resetting reverse_key if needed
                if reverse_key and direction is RelationType.ONE_TO_ONE and \
                        hasattr(child_merged_model, reverse_key) and \
                        getattr(child_merged_model, reverse_key) != result_model:
                    setattr(child_merged_model, reverse_key, None)
                    db_adapter.save_model(child_merged_model)
                continue

            if child_merged_model != this_child_model:
                if fkey in merge_cls_policy:
                    merge_function = merge_cls_policy[fkey]
                    if merge_function is False:
                        raise MergeModelsNotAllowed(this_model, merged_model,
                                                    fkey)
                    
                    result_child_model = merge_function(
                        this_model, merged_model,
                        result_child_model, child_merged_model,
                        fkey, reverse_key)
                else:
                    raise CannotMergeModels(fkey,
                                            this_model, this_child_model,
                                            merged_model, child_merged_model)
            
        # saving f_model
        setattr(result_model, fkey, result_child_model)
            

    # combining x-to-many models
    for fkey, direction, reverse_key in relations:
        if fkey in ignore_fields or not direction.is_x_to_many():
            continue
        
        result_children = None
        for other_model in models:
            other_children = list(db_adapter.get_related_x_to_many(other_model,
                                                                   fkey))
            
            if other_children and direction is RelationType.ONE_TO_MANY and \
                    fkey in merge_cls_policy:

                if result_children is None:
                    result_children = list(
                        db_adapter.get_related_x_to_many(result_model, fkey))
                if result_children:
                    merge_function = merge_cls_policy[fkey]
                    __one_to_many_merge(db_adapter,
                                        result_model, other_model,
                                        result_children, other_children,
                                        fkey, reverse_key,
                                        merge_function,
                                        ignore_fields=ignore_fields)

            db_adapter.add_related_models(result_model, fkey, other_children)

            
    
    # grabbing plain values from other models
    for key in fields:
        if key in ignore_fields or hasattr(result_model, key) and \
                getattr(result_model, key) is not None:
            continue
        # looking for first value to grab
        for model in models:
            if hasattr(model, key) and getattr(model, key) is not None:
                setattr(result_model, key, getattr(model, key))
                break
            
    # deleting all merged models
    for model in models:
        db_adapter.delete(model)
    
    return result_model


import sqlalchemy as sa
from sqlalchemy.orm import relationship

from ..config import Base



class __CommonColumns(object):
    f_binary = sa.Column(sa.LargeBinary, nullable=True)
    f_boolean = sa.Column(sa.Boolean, nullable=True)
    f_string = sa.Column(sa.String, nullable=True)
    f_text = sa.Column(sa.Text, nullable=True)
    f_integer = sa.Column(sa.Integer, nullable=True)
    f_float = sa.Column(sa.Float, nullable=True)
    f_date = sa.Column(sa.Date, nullable=True)
    f_time = sa.Column(sa.Time, nullable=True)
    f_datetime = sa.Column(sa.DateTime, nullable=True)


class ModelManyRefsOne(__CommonColumns, Base):
    __tablename__ = 'model_many_refs_one'
    
    id = sa.Column(sa.Integer, primary_key=True)
    
    # many-to-one
    two_x_1_id_a = sa.Column(
        sa.Integer,
        sa.ForeignKey('model_many_refs_two.id', ondelete='CASCADE',
                      use_alter=True))
    two_x_1_a = relationship(
        'ModelManyRefsTwo',
        back_populates='one_1_x_a',
        foreign_keys=[two_x_1_id_a],
        post_update=True)
    
    # one-to-many
    two_1_x_b = relationship(
        'ModelManyRefsTwo',
        back_populates='one_x_1_b',
        foreign_keys='[ModelManyRefsTwo.one_x_1_id_b]',
        post_update=True)
    
    # many-to-many
    two_x_x_a = relationship(
        'ModelManyRefsTwo',
        secondary='model_many_refs_one_to_two_a',
        back_populates='one_x_x_a',
        post_update=True)
    
    two_x_x_b = relationship(
        'ModelManyRefsTwo',
        secondary='model_many_refs_one_to_two_b',
        back_populates='one_x_x_b',
        post_update=True)
    
    # one-to-one
    two_1_1_id_a = sa.Column(
        sa.Integer,
        sa.ForeignKey('model_many_refs_two.id', ondelete='CASCADE',
                      use_alter=True),
        unique=True)
    two_1_1_a = relationship(
        'ModelManyRefsTwo',
        back_populates='one_1_1_a',
        foreign_keys=[two_1_1_id_a],
        post_update=True)
    
    two_1_1_b = relationship(
        'ModelManyRefsTwo',
        back_populates='one_1_1_b',
        foreign_keys='[ModelManyRefsTwo.one_1_1_id_b]',
        uselist=False,
        post_update=True)
    

class ModelManyRefsOneToTwoA(Base):
    __tablename__ = 'model_many_refs_one_to_two_a'
    
    one_id = sa.Column(
        sa.Integer,
        sa.ForeignKey(ModelManyRefsOne.id, ondelete='CASCADE'),
        primary_key=True)
    two_id = sa.Column(
        sa.Integer,
        sa.ForeignKey('model_many_refs_two.id', ondelete='CASCADE'),
        primary_key=True)


class ModelManyRefsOneToTwoB(Base):
    __tablename__ = 'model_many_refs_one_to_two_b'
    
    one_id = sa.Column(
        sa.Integer,
        sa.ForeignKey(ModelManyRefsOne.id, ondelete='CASCADE',
                      use_alter=True),
        primary_key=True)
    two_id = sa.Column(
        sa.Integer,
        sa.ForeignKey('model_many_refs_two.id', ondelete='CASCADE',
                      use_alter=True),
        primary_key=True)


class ModelManyRefsTwo(__CommonColumns, Base):
    __tablename__ = 'model_many_refs_two'
    
    id = sa.Column(sa.Integer, primary_key=True)
    
    # one-to-many
    one_1_x_a = relationship(
        ModelManyRefsOne,
        back_populates='two_x_1_a',
        foreign_keys='[ModelManyRefsOne.two_x_1_id_a]',
        post_update=True)
    
    # many-to-one
    one_x_1_id_b = sa.Column(
        sa.Integer,
        sa.ForeignKey('model_many_refs_one.id', ondelete='CASCADE'))
    one_x_1_b = relationship(
        ModelManyRefsOne,
        back_populates='two_1_x_b',
        foreign_keys=[one_x_1_id_b],
        post_update=True)
    
    # many-to-many
    one_x_x_a = relationship(
        ModelManyRefsOne,
        secondary='model_many_refs_one_to_two_a',
        back_populates='two_x_x_a',
        post_update=True)
    
    one_x_x_b = relationship(
        ModelManyRefsOne,
        secondary='model_many_refs_one_to_two_b',
        back_populates='two_x_x_b',
        post_update=True)
    
    # one-to-one
    one_1_1_a = relationship(
        ModelManyRefsOne,
        back_populates='two_1_1_a',
        foreign_keys=[ModelManyRefsOne.two_1_1_id_a],
        uselist=False,
        post_update=True)
    
    one_1_1_id_b = sa.Column(
        sa.Integer,
        sa.ForeignKey('model_many_refs_one.id', ondelete='CASCADE'),
        unique=True)
    one_1_1_b = relationship(
        ModelManyRefsOne,
        back_populates='two_1_1_b',
        foreign_keys=[one_1_1_id_b],
        post_update=True)


from itertools import chain

from save_to_db.core.item import Item
from save_to_db.core.scope import Scope
from save_to_db.utils.test_base import TestBase



class TestModelDeleter(TestBase):
    
    ModelGeneralOne = None
    ModelGeneralTwo = None
    
    
    def test_model_deleter(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = [{'f_integer', 'f_string'}]
            getters = [{'f_integer', 'f_string'}]
            deleter_selectors = ['f_string']
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
        
        persister = self.persister
        get_key = lambda model: (model.f_string, model.f_integer)
        
        # imitating first collection from different sources
        for source_number in range(1, 4):
            for outer_i in range(1, 3):
                bulk = ItemGeneralOne.Bulk(
                    f_string='source-{}'.format(source_number))
                for inner_i in range(1, 3):
                    bulk.gen(f_integer=outer_i*100 + inner_i)
                persister.persist(bulk)
        
        # will delete nothing
        persister.execute_deleter(ItemGeneralOne)
        model_ones = self.get_all_models(self.ModelGeneralOne,
                                         sort_key=get_key)

        model_keys = [get_key(model) for model in model_ones]
        expected = [
            # source-1 first iteration
            ('source-1', 101),
            ('source-1', 102),
            # source-1 second iteration
            ('source-1', 201),
            ('source-1', 202),
            # source-2 first iteration
            ('source-2', 101),
            ('source-2', 102),
            # source-2 second iteration
            ('source-2', 201),
            ('source-2', 202),
            # source-3 first iteration
            ('source-3', 101),
            ('source-3', 102),
            # source-3 second iteration
            ('source-3', 201),
            ('source-3', 202),
        ]
        self.assertEqual(model_keys, expected)
        
        # imitating first collection from different sources
        for source_number in range(2, 4):
            for outer_i in range(2, 3):
                bulk = ItemGeneralOne.Bulk(
                    f_string='source-{}'.format(source_number))
                for inner_i in range(1, 4):
                    bulk.gen(f_integer=outer_i*100 + inner_i)
                persister.persist(bulk)
        
        persister.execute_deleter(ItemGeneralOne)
        model_ones = self.get_all_models(self.ModelGeneralOne,
                                         sort_key=get_key)

        model_keys = [get_key(model) for model in model_ones]
        expected = [
            # source-1 must have been ignored (and not deleted)
            ('source-1', 101),
            ('source-1', 102),
            ('source-1', 201),
            ('source-1', 202),
            # source-2 starts from second iteration
            ('source-2', 201),
            ('source-2', 202),
            ('source-2', 203),  # new model
            # source-3 starts from second iteration
            ('source-3', 201),
            ('source-3', 202),
            ('source-3', 203),  # new model
        ]
        self.assertEqual(model_keys, expected)
        
        
    def test_model_deleter_execute_on_persist(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = [{'f_integer', 'f_string'}]
            getters = [{'f_integer', 'f_string'}]
            deleter_selectors = ['f_string']
            deleter_execute_on_persist = True
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            creators = [{'f_integer', 'f_string'}]
            getters = [{'f_integer', 'f_string'}]
            deleter_selectors = ['f_string']
        
        persister = self.persister
        get_key = lambda model: (model.f_string, model.f_integer)
        
        # creating items
        item_one_bulk = ItemGeneralOne.Bulk()
        item_two_bulk = ItemGeneralTwo.Bulk()
        for i in range(1, 3):
            item_one_bulk.gen(f_integer=100+i, f_string='source-1')
            item_two_bulk.gen(f_integer=100+i, f_string='source-1')
        
        persister.persist(item_one_bulk)
        persister.persist(item_two_bulk)
        
        # recreating items (first item from bulk one must be deleted)
        item_one_bulk = ItemGeneralOne.Bulk()
        item_two_bulk = ItemGeneralTwo.Bulk()
        for i in range(2, 4):
            item_one_bulk.gen(f_integer=100+i, f_string='source-1')
            item_two_bulk.gen(f_integer=100+i, f_string='source-1')
        
        persister.persist(item_one_bulk)
        persister.persist(item_two_bulk)
        
        models_one = self.get_all_models(self.ModelGeneralOne,
                                         sort_key=get_key)
        model_one_keys = [get_key(model) for model in models_one]
        expect = [('source-1', 102), ('source-1', 103)]
        self.assertEqual(model_one_keys, expect)
        
        models_two = self.get_all_models(self.ModelGeneralTwo,
                                         sort_key=get_key)
        model_two_keys = [get_key(model) for model in models_two]
        expect = [('source-1', 101), ('source-1', 102), ('source-1', 103)]
        self.assertEqual(model_two_keys, expect)
    
    
    def test_model_deleter_execute_scoped(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = [{'f_integer', 'f_string'}]
            getters = [{'f_integer', 'f_string'}]
            deleter_selectors = ['f_string']
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
        
        persister = self.persister
        get_key = lambda model: (model.f_string, model.f_integer)
        
        
        test_scope = Scope(
            {}, collection_id='test_model_deleter_execute_scoped_dummy')
        persister.persist(test_scope[ItemGeneralOne](
            f_string='source-9000', f_integer='9000'
        ))
        
        
        # imitating first collection from different sources
        for source_number in range(1, 3):
            bulk = ItemGeneralOne.Bulk(
                f_string='source-{}'.format(source_number))
            for inner_i in range(1, 3):
                bulk.gen(f_integer=100 + inner_i)
            persister.persist(bulk)
        
        # will delete nothing (db is empty)
        self.assertEqual(
            ItemGeneralOne.metadata['model_deleter'].selectors,
            [{'f_string': 'source-1'}, {'f_string': 'source-2'}],
        )
        self.assertEqual(
            ItemGeneralOne.metadata['model_deleter'].keepers,
            [{'id': 2}, {'id': 3}, {'id': 4}, {'id': 5}],
        )
        persister.execute_scope_deleter(ItemGeneralOne.get_collection_id())
        
        # delete must be reset after executing
        self.assertFalse(ItemGeneralOne.metadata['model_deleter'].selectors)
        self.assertFalse(ItemGeneralOne.metadata['model_deleter'].keepers)
        
        model_ones = self.get_all_models(self.ModelGeneralOne,
                                         sort_key=get_key)

        model_keys = [get_key(model) for model in model_ones]
        expected = [
            # source-1
            ('source-1', 101),
            ('source-1', 102),
            # source-2
            ('source-2', 101),
            ('source-2', 102),
            # from test scope
            ('source-9000', 9000),
        ]
        self.assertEqual(model_keys, expected)
        
        
        # delete without scope
        # ('source-1', 101) and ('source-2', 101) must be deleted
        scope = Scope({}, collection_id='test_model_deleter_execute_scoped')
        ScopedItemGeneralOne = scope[ItemGeneralOne]
        
        for source_number in range(1, 3):
            bulk = ItemGeneralOne.Bulk(
                f_string='source-{}'.format(source_number))
            for inner_i in range(2, 4):
                bulk.gen(f_integer=100 + inner_i)
            persister.persist(bulk)
        
        persister.execute_scope_deleter(ItemGeneralOne.get_collection_id())
        model_ones = self.get_all_models(self.ModelGeneralOne,
                                         sort_key=get_key)

        model_keys = [get_key(model) for model in model_ones]
        expected = [
            # source-1
            ('source-1', 102),
            ('source-1', 103),
            # source-2
            ('source-2', 102),
            ('source-2', 103),
            # from test scope
            ('source-9000', 9000),
        ]
        self.assertEqual(model_keys, expected)
        
        # delete with scope
        persisted_models = []
        for source_number in range(1, 3):
            bulk = ScopedItemGeneralOne.Bulk(
                f_string='source-{}'.format(source_number))
            for inner_i in range(3, 5):
                bulk.gen(f_integer=100 + inner_i)
            _, models_list = persister.persist(bulk)
            for model_list in models_list:
                persisted_models.extend(model_list)
        
        # before executing making sure that different model deleter was used
        self.assertFalse(ItemGeneralOne.metadata['model_deleter'].selectors)
        self.assertFalse(ItemGeneralOne.metadata['model_deleter'].keepers)
        
        expect_selectors = [
            {'f_string': 'source-1'},
            {'f_string': 'source-2'},
        ]
        scoped_deleter = ScopedItemGeneralOne.metadata['model_deleter']
        self.assertEqual(scoped_deleter.selectors, expect_selectors)
        
        persisted_models.sort(key=lambda model: model.id)
        scoped_deleter.keepers.sort(key=lambda keeper: keeper['id'])
        expect_keepers = [{'id': model.id} for model in persisted_models]
        self.assertTrue(len(expect_keepers) != 0)
        self.assertEqual(scoped_deleter.keepers, expect_keepers)
        
        # creating model without scope
        persister.persist(ItemGeneralOne(f_string='source-9009',
                                         f_integer='9009'))
        
        persister.execute_scope_deleter(
            ScopedItemGeneralOne.get_collection_id()
        )
        model_ones = self.get_all_models(self.ModelGeneralOne,
                                         sort_key=get_key)

        model_keys = [get_key(model) for model in model_ones]
        expected = [
            # source-1
            ('source-1', 103),
            ('source-1', 104),
            # source-2
            ('source-2', 103),
            ('source-2', 104),
            # from test scope
            ('source-9000', 9000),
            # from not scoped item class with deleter not executed
            ('source-9009', 9009),
        ]
        self.assertEqual(model_keys, expected)
    
    
    def test_model_unrefs(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = [{'f_integer'}]
            getters = [{'f_integer'}]
            unref_x_to_many = {
                'two_1_x': ['f_string'],
                'two_x_x': ['f_string']
            }
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            creators = [{'f_integer', 'f_string'}]
            getters = [{'f_integer', 'f_string'}]
        
        persister = self.persister
        get_key = lambda model: (model.f_text, model.f_string, model.f_integer)
        
        item_one = ItemGeneralOne(f_integer=100)
        for source_name in ['src-1', 'src-2']:
            for i in range(1, 3):
                item_one['two_1_x'].gen(f_integer=200+i, f_string=source_name,
                                        f_text='two_1_x')
                item_one['two_x_x'].gen(f_integer=900+i, f_string=source_name,
                                        f_text='two_x_x')
        persister.persist(item_one)
        
        model_ones = self.get_all_models(self.ModelGeneralOne)
        self.assertEqual(len(model_ones), 1)
        model_one = model_ones[0]
        self.assertEqual(model_one.f_integer, 100)
        two_1_x = self.get_related_x_to_many(model_one, 'two_1_x',
                                             sort_key=get_key)
        two_x_x = self.get_related_x_to_many(model_one, 'two_x_x',
                                             sort_key=get_key)
        self.assertEqual(len(two_1_x), 4)
        self.assertEqual(len(two_x_x), 4)
        keys = [get_key(model) for model in chain(two_1_x, two_x_x)]
        expect = [
            ('two_1_x', 'src-1', 201),
            ('two_1_x', 'src-1', 202),
            ('two_1_x', 'src-2', 201),
            ('two_1_x', 'src-2', 202),
            ('two_x_x', 'src-1', 901),
            ('two_x_x', 'src-1', 902),
            ('two_x_x', 'src-2', 901),
            ('two_x_x', 'src-2', 902)
        ]
        self.assertEqual(keys, expect)
        
        item_one = ItemGeneralOne(f_integer=100)
        for source_name in ['src-1', 'src-2']:
            for i in range(2, 4):
                item_one['two_1_x'].gen(f_integer=200+i, f_string=source_name,
                                        f_text='two_1_x')
                item_one['two_x_x'].gen(f_integer=900+i, f_string=source_name,
                                        f_text='two_x_x')
        persister.persist(item_one)
        model_ones = self.get_all_models(self.ModelGeneralOne)
        self.assertEqual(len(model_ones), 1)
        model_one = model_ones[0]
        self.assertEqual(model_one.f_integer, 100)
        two_1_x = self.get_related_x_to_many(model_one, 'two_1_x',
                                             sort_key=get_key)
        two_x_x = self.get_related_x_to_many(model_one, 'two_x_x',
                                             sort_key=get_key)
        self.assertEqual(len(two_1_x), 4)
        self.assertEqual(len(two_x_x), 4)
        keys = [get_key(model) for model in chain(two_1_x, two_x_x)]
        # f_integer 201 and 901 were removed
        expect = [
            ('two_1_x', 'src-1', 202),
            ('two_1_x', 'src-1', 203),
            ('two_1_x', 'src-2', 202),
            ('two_1_x', 'src-2', 203),
            ('two_x_x', 'src-1', 902),
            ('two_x_x', 'src-1', 903),
            ('two_x_x', 'src-2', 902),
            ('two_x_x', 'src-2', 903)
        ]
        self.assertEqual(keys, expect)
        
        # f_integer 201 and 901 are still in database
        model_twos = self.get_all_models(self.ModelGeneralTwo,
                                         sort_key=get_key)
        keys = [get_key(model) for model in model_twos]
        expect = [
            ('two_1_x', 'src-1', 201),
            ('two_1_x', 'src-1', 202),
            ('two_1_x', 'src-1', 203),
            ('two_1_x', 'src-2', 201),
            ('two_1_x', 'src-2', 202),
            ('two_1_x', 'src-2', 203),
            ('two_x_x', 'src-1', 901),
            ('two_x_x', 'src-1', 902),
            ('two_x_x', 'src-1', 903),
            ('two_x_x', 'src-2', 901),
            ('two_x_x', 'src-2', 902),
            ('two_x_x', 'src-2', 903)
        ]
        self.assertEqual(keys, expect)
        
        # delete must be reset after executing
        self.assertFalse(
            ItemGeneralOne.metadata['model_unrefs']['two_1_x'].selectors)
        self.assertFalse(
            ItemGeneralOne.metadata['model_unrefs']['two_1_x'].keepers)
        self.assertFalse(
            ItemGeneralOne.metadata['model_unrefs']['two_x_x'].selectors)
        self.assertFalse(
            ItemGeneralOne.metadata['model_unrefs']['two_x_x'].keepers)